#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from typing import Dict, Any

from .component_state import ComponentState, L, P, SCOPE_TO_STATE_MAP
from ..pdi.d4_req import D4Req
from ..protocol.constants import CommandScope
from ..pdi.constants import PdiCommand, D4Action


class BaseState(ComponentState):
    """
    Maintain the state of a Lionel Base
    """

    def __init__(self, scope: CommandScope = CommandScope.BASE) -> None:
        if scope != CommandScope.BASE:
            raise ValueError(f"Invalid scope: {scope}")
        super().__init__(scope)
        self._base_name = None
        self._firmware = None
        self._firmware_high = None
        self._firmware_low = None
        self._route_throw_rate = None
        self._d4_engines: int | None = None
        self._d4_trains: int | None = None
        self._first_d4_engine_rec_no: int | None = None
        self._first_d4_train_rec_no: int | None = None

    def __repr__(self) -> str:
        bn = f"Lionel Base 3: {self._base_name if self._base_name else 'NA'}"
        fw = f" Firmware: {self._firmware if self._firmware else 'NA'}"
        d4e = f" 4-digit Engines: {self._d4_engines if self._d4_engines is not None else 'NA'}"
        if d4e != "NA" and self._first_d4_engine_rec_no is not None:
            d4e += f" (FRN: {self._first_d4_engine_rec_no})"
        d4t = f" 4-digit Trains: {self._d4_trains if self._d4_trains is not None else 'NA'}"
        if d4t != "NA" and self._first_d4_train_rec_no is not None:
            d4t += f" (FRN: {self._first_d4_train_rec_no})"
        return f"{bn}{fw}{d4e}{d4t}"

    def update(self, command: L | P) -> None:
        from ..pdi.base_req import BaseReq
        from ..pdi.d4_req import D4Req

        if isinstance(command, BaseReq):
            with self._cv:
                # Note: super().update is explicitly not called
                self._base_name = command.name.title() if command.name else self._base_name
                self._firmware = command.firmware if command.firmware else self._firmware
                self._is_known = True
                if self.firmware:
                    version_info = self.firmware.split(".")
                    self._firmware_high = int(version_info[0])
                    self._firmware_low = int(version_info[1])
                self._route_throw_rate = command.route_throw_rate
                self.changed.set()
                self._cv.notify_all()
        elif isinstance(command, D4Req):
            with self._cv:
                if command.pdi_command == PdiCommand.D4_ENGINE:
                    if command.action == D4Action.COUNT:
                        self._d4_engines = command.count
                    elif command.action == D4Action.FIRST_REC:
                        self._first_d4_engine_rec_no = command.record_no if command.record_no != 0xFFFF else None
                elif command.pdi_command == PdiCommand.D4_TRAIN:
                    if command.action == D4Action.COUNT:
                        self._d4_trains = command.count
                    elif command.action == D4Action.FIRST_REC:
                        self._first_d4_train_rec_no = command.record_no if command.record_no != 0xFFFF else None
            # do not signal state update for engine and train counts

    @property
    def base_name(self) -> str:
        return self._base_name

    @property
    def firmware(self) -> str:
        return self._firmware

    @property
    def firmware_high(self) -> int:
        return self._firmware_high

    @property
    def firmware_low(self) -> int:
        return self._firmware_low

    @property
    def d4_engines(self) -> int:
        return self._d4_engines

    @property
    def first_d4_engine_record_no(self) -> int:
        return self._first_d4_engine_rec_no

    @property
    def d4_trains(self) -> int:
        return self._d4_trains

    @property
    def first_d4_train_record_no(self) -> int:
        return self._first_d4_train_rec_no

    @property
    def route_throw_rate(self) -> float:
        return self._route_throw_rate

    @property
    def is_tmcc(self) -> bool:
        return False

    @property
    def is_legacy(self) -> bool:
        return True

    @property
    def is_lcs(self) -> bool:
        return True

    def as_bytes(self) -> bytes:
        if self.is_known:
            from ..pdi.base_req import BaseReq

            byte_str = BaseReq(self.address, PdiCommand.BASE, state=self).as_bytes
            if self.d4_engines is not None:
                byte_str += D4Req(0, PdiCommand.D4_ENGINE, action=D4Action.COUNT, count=self.d4_engines).as_bytes
            if self.first_d4_engine_record_no is not None:
                byte_str += D4Req(
                    self.first_d4_engine_record_no, PdiCommand.D4_ENGINE, action=D4Action.FIRST_REC
                ).as_bytes
            if self.d4_trains is not None:
                byte_str += D4Req(0, PdiCommand.D4_TRAIN, action=D4Action.COUNT, count=self.d4_trains).as_bytes
            if self.first_d4_train_record_no is not None:
                byte_str += D4Req(
                    self.first_d4_train_record_no, PdiCommand.D4_TRAIN, action=D4Action.FIRST_REC
                ).as_bytes
            return byte_str
        else:
            return bytes()

    def as_dict(self) -> Dict[str, Any]:
        d = dict()
        d["firmware"] = self.firmware
        d["base_name"] = self.base_name
        d["route_throw_rate"] = self.route_throw_rate
        return d


SCOPE_TO_STATE_MAP.update({CommandScope.BASE: BaseState})



from __future__ import annotations

import abc
from abc import ABC
from threading import Thread
from time import sleep
from typing import Protocol, TypeVar, Generic

from gpiozero import LED, PingServer

from ..db.component_state import ComponentState
from ..db.component_state_store import ComponentStateStore
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum, TMCC1SwitchCommandEnum

T = TypeVar("T", bound=ComponentState)


class StateValidator(Protocol):
    def __call__(self, state: StateSource) -> bool: ...


# noinspection PyUnresolvedReferences
class StateSource(ABC, Thread, Generic[T]):
    __metaclass__ = abc.ABCMeta

    def __init__(self, scope: CommandScope, address: int, active_led: LED, inactive_led: LED = None) -> None:
        super().__init__(daemon=True, name=f"{scope.label} {address} State")
        self._scope = scope
        self._address = address
        self.active_led = active_led
        self.inactive_led = inactive_led
        self._component: T = ComponentStateStore.build().component(scope, address)
        self._is_running = True
        self.start()

    def reset(self) -> None:
        self._is_running = False
        if self._component:
            with self._component.synchronizer:
                self._component.synchronizer.notify_all()

    def run(self) -> None:
        while self._is_running:
            with self._component.synchronizer:
                self._component.synchronizer.wait()
                if self._is_running:
                    self.active_led.value = 1 if self.is_active is True else 0
                    if self.inactive_led:
                        self.inactive_led.value = 0 if self.is_active is True else 1

    @property
    @abc.abstractmethod
    def is_active(self) -> bool: ...


class LionelBaseSource(StateSource):
    def __init__(
        self,
        server: str,
        led: LED,
        delay: float = 10,
    ) -> None:
        self._delay = delay
        self._pieces = server.split(".")
        if len(self._pieces) == 4:
            address = int(self._pieces[3])
        else:
            address = 99
        self._ping = PingServer(server, event_delay=delay)
        super().__init__(CommandScope.SYSTEM, address, led)

    def __iter__(self):
        return self

    def __next__(self):
        return self.is_active

    @property
    def is_active(self) -> bool:
        return self._ping.is_active

    def run(self) -> None:
        while self._is_running:
            self.active_led.value = 1 if self.is_active else 0
            sleep(self._delay)


class SwitchStateSource(StateSource):
    def __init__(self, address: int, thru_led: LED, out_led: LED) -> None:
        self._state = TMCC1SwitchCommandEnum.THRU
        super().__init__(CommandScope.SWITCH, address, thru_led, out_led)

    def __iter__(self):
        return self

    def __next__(self):
        return self.is_active

    @property
    def is_active(self) -> bool:
        return self._component.state == self._state


class RouteStateSource(StateSource):
    def __init__(self, address: int, active_led: LED, inactive_led: LED) -> None:
        self._state = TMCC1SwitchCommandEnum.THRU
        super().__init__(CommandScope.ROUTE, address, active_led, inactive_led)

    def __iter__(self):
        return self

    def __next__(self):
        return self.is_active

    @property
    def is_active(self) -> bool:
        return self._component.is_active


class AccessoryStateSource(StateSource):
    def __init__(
        self,
        address: int,
        led: LED,
        aux_state: TMCC1AuxCommandEnum = None,
        aux1_state: TMCC1AuxCommandEnum = None,
        aux2_state: TMCC1AuxCommandEnum = None,
    ) -> None:
        self._aux_state = aux_state
        self._aux1_state = aux1_state
        self._aux2_state = aux2_state
        super().__init__(CommandScope.ACC, address, led)

    def __iter__(self):
        return self

    def __next__(self):
        return self.is_active

    @property
    def is_active(self) -> bool:
        return (
            (self._aux_state is not None and self._component.aux_state == self._aux_state)
            or (self._aux1_state is not None and self._component.aux1_state == self._aux1_state)
            or (self._aux2_state is not None and self._component.aux2_state == self._aux2_state)
        )


class EngineStateSource(StateSource):
    def __init__(
        self,
        address: int,
        led: LED,
        func: StateValidator = None,
        scope: CommandScope = CommandScope.ENGINE,
    ) -> None:
        self._func = func
        self._scope = scope
        super().__init__(scope, address, led)

    def __iter__(self):
        return self

    def __next__(self):
        return self.is_active

    @property
    def is_active(self) -> bool:
        return self._func(self._component) if self._func else False



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from __future__ import annotations

from collections import deque
from threading import RLock
from typing import TypeVar, Generic, Iterable

_T = TypeVar("_T")


class UniqueDeque(deque, Generic[_T]):
    def __init__(self, iterable: Iterable[_T] = None, maxlen: int = None) -> None:
        self._lock = RLock()
        super().__init__(maxlen=maxlen)
        self._seen: set[_T] = set()
        if iterable is not None:
            self.extend(iterable)

    def clear(self):
        with self._lock:
            self._seen.clear()
            super().clear()

    def __add__(self, value: UniqueDeque[_T] | deque[_T] | list[_T] | tuple[_T], /) -> UniqueDeque[_T]:
        with self._lock:
            uq = self.copy()
            for x in value:
                uq.append(x)
            return uq

    def __iadd__(self, value: UniqueDeque[_T] | deque[_T] | list[_T] | tuple[_T], /) -> UniqueDeque[_T]:
        with self._lock:
            for x in value:
                self.append(x)
            return self

    def __delitem__(self, key, /):
        with self._lock:
            super().__delitem__(key)
            self._seen.discard(key)

    def __setitem__(self, key, value: _T, /):
        with self._lock:
            if value not in self._seen:
                self._seen.add(value)
                super().__setitem__(key, value)

    def __reduce__(self):
        raise NotImplementedError

    def __imul__(self, value, /):
        raise NotImplementedError

    def __mul__(self, value, /):
        raise NotImplementedError

    def insert(self, i, x: _T, /):
        raise NotImplementedError

    def remove(self, value: _T, /):
        with self._lock:
            super().remove(value)
            self._seen.discard(value)

    def popleft(self) -> _T:
        with self._lock:
            val = super().popleft()
            self._seen.discard(val)
            return val

    def pop(self):
        with self._lock:
            val = super().pop()
            self._seen.discard(val)
            return val

    def extendleft(self, iterable: Iterable[_T], /):
        with self._lock:
            for x in iterable:
                self.appendleft(x)

    def extend(self, iterable: Iterable[_T], /) -> None:
        with self._lock:
            for x in iterable:
                self.append(x)

    def copy(self) -> UniqueDeque[_T]:
        with self._lock:
            uq = UniqueDeque(self, self.maxlen)
            return uq

    def appendleft(self, x: _T, /) -> None:
        with self._lock:
            if x in self._seen:
                if self[0] != x:
                    super().remove(x)
                    super().appendleft(x)
            else:
                if self.maxlen is not None and len(self) == self.maxlen:
                    self.pop()
                self._seen.add(x)
                super().appendleft(x)

    def push(self, x: _T, /):
        self.appendleft(x)

    def append(self, x: _T, /) -> None:
        with self._lock:
            if x in self._seen:
                if self[-1] != x:
                    super().remove(x)
                    super().append(x)
            else:
                if self.maxlen is not None and len(self) == self.maxlen:
                    self.popleft()
                self._seen.add(x)
                super().append(x)



from __future__ import annotations

import abc
from enum import IntEnum, unique
from typing import List

from ..command_def import Mixins
from ..tmcc2.tmcc2_constants import (
    TMCC2_SOUND_OFF_COMMAND,
    TMCC2_SOUND_ON_COMMAND,
    TMCC2CommandDef,
    TMCC2CommandPrefix,
    TMCC2Enum,
)

"""
    Legacy/TMCC2 Multi-byte Command sequences
"""


class TMCC2MultiByteEnum(TMCC2Enum):
    """
    Marker Interface for all TMCC2 enums
    """

    pass


"""
    Word #1 - Function indexes
"""
TMCC2_R4LC_INDEX_PREFIX: int = 0x40
TMCC2_VARIABLE_INDEX_PREFIX: int = 0x60
TMCC2_PARAMETER_INDEX_PREFIX: int = 0x70


@unique
class TMCCPrefixEnum(Mixins, IntEnum):
    R4LC = TMCC2_R4LC_INDEX_PREFIX
    VARIABLE = TMCC2_VARIABLE_INDEX_PREFIX
    PARAMETER = TMCC2_PARAMETER_INDEX_PREFIX


# Parameter indexes
TMCC2_PARAMETER_ASSIGNMENT_PARAMETER_INDEX: int = 0x71
TMCC2_RAIL_SOUNDS_DIALOG_TRIGGERS_PARAMETER_INDEX: int = 0x72
TMCC2_RAIL_SOUNDS_EFFECTS_TRIGGERS_PARAMETER_INDEX: int = 0x74
TMCC2_RAIL_SOUNDS_MASKING_CONTROL_PARAMETER_INDEX: int = 0x76
TMCC2_EFFECTS_CONTROLS_PARAMETER_INDEX: int = 0x7C
TMCC2_LIGHTING_CONTROLS_PARAMETER_INDEX: int = 0x7D
TMCC2_VARIABLE_LENGTH_COMMAND_PARAMETER_INDEX: int = 0x7F


@unique
class TMCC2ParameterIndex(Mixins, IntEnum):
    PARAMETER_ASSIGNMENT = TMCC2_PARAMETER_ASSIGNMENT_PARAMETER_INDEX
    DIALOG_TRIGGERS = TMCC2_RAIL_SOUNDS_DIALOG_TRIGGERS_PARAMETER_INDEX
    EFFECTS_TRIGGERS = TMCC2_RAIL_SOUNDS_EFFECTS_TRIGGERS_PARAMETER_INDEX
    MASKING_CONTROLS = TMCC2_RAIL_SOUNDS_MASKING_CONTROL_PARAMETER_INDEX
    EFFECTS_CONTROLS = TMCC2_EFFECTS_CONTROLS_PARAMETER_INDEX
    LIGHTING_CONTROLS = TMCC2_LIGHTING_CONTROLS_PARAMETER_INDEX
    VARIABLE_LENGTH_COMMAND = TMCC2_VARIABLE_LENGTH_COMMAND_PARAMETER_INDEX


class MultiByteCommandDef(TMCC2CommandDef):
    __metaclass__ = abc.ABCMeta

    def __init__(self, command_bits: int, d_min: int = 0, d_max: int = 0, interval: int = None) -> None:
        super().__init__(command_bits, d_min=d_min, d_max=d_max, interval=interval)
        self._first_byte = TMCC2CommandPrefix.ENGINE


class TMCC2ParameterEnum(TMCC2MultiByteEnum):
    """
    Marker interface for all Parameter Data enums
    """

    pass


"""
    Word #2 - RailSounds Dialog trigger controls (index 0x2)
"""
TMCC2_DIALOG_CONTROL_CONVENTIONAL_SHUTDOWN: int = 0x01
TMCC2_DIALOG_CONTROL_SCENE_TWO: int = 0x02
TMCC2_DIALOG_CONTROL_SCENE_SEVEN: int = 0x03
TMCC2_DIALOG_CONTROL_SCENE_FIVE: int = 0x04
TMCC2_DIALOG_CONTROL_SHORT_HORN: int = 0x05
TMCC2_DIALOG_CONTROL_TOWER_ENGINE_STARTUP: int = 0x06
TMCC2_DIALOG_CONTROL_ENGINEER_DEPARTURE_DENIED: int = 0x07
TMCC2_DIALOG_CONTROL_ENGINEER_DEPARTURE_GRANTED: int = 0x08
TMCC2_DIALOG_CONTROL_ENGINEER_HAVE_DEPARTED: int = 0x09
TMCC2_DIALOG_CONTROL_ENGINEER_ALL_CLEAR: int = 0x0A
TMCC2_DIALOG_CONTROL_TOWER_STOP_HOLD: int = 0x0B
TMCC2_DIALOG_CONTROL_TOWER_RESTRICTED_SPEED: int = 0x0C
TMCC2_DIALOG_CONTROL_TOWER_SLOW_SPEED: int = 0x0D
TMCC2_DIALOG_CONTROL_TOWER_MEDIUM_SPEED: int = 0x0E
TMCC2_DIALOG_CONTROL_TOWER_LIMITED_SPEED: int = 0x0F
TMCC2_DIALOG_CONTROL_TOWER_NORMAL_SPEED: int = 0x10
TMCC2_DIALOG_CONTROL_TOWER_HIGHBALL_SPEED: int = 0x11
TMCC2_DIALOG_CONTROL_ENGINEER_ARRIVING_SOON: int = 0x12
TMCC2_DIALOG_CONTROL_ENGINEER_HAVE_ARRIVED: int = 0x13
TMCC2_DIALOG_CONTROL_ENGINEER_SHUTDOWN: int = 0x14
TMCC2_DIALOG_CONTROL_ENGINEER_ID: int = 0x15
TMCC2_DIALOG_CONTROL_ENGINEER_ACK: int = 0x16
TMCC2_DIALOG_CONTROL_ENGINEER_STOP_SPEED_ACK: int = 0x17
TMCC2_DIALOG_CONTROL_ENGINEER_RESTRICTED_SPEED_ACK: int = 0x18
TMCC2_DIALOG_CONTROL_ENGINEER_SLOW_SPEED_ACK: int = 0x19
TMCC2_DIALOG_CONTROL_ENGINEER_MEDIUM_SPEED_ACK: int = 0x1A
TMCC2_DIALOG_CONTROL_ENGINEER_LIMITED_SPEED_ACK: int = 0x1B
TMCC2_DIALOG_CONTROL_ENGINEER_NORMAL_SPEED_ACK: int = 0x1C
TMCC2_DIALOG_CONTROL_ENGINEER_HIGHBALL_SPEED_ACK: int = 0x1D

TMCC2_DIALOG_CONTROL_ENGINEER_CONTEXT_DEPENDENT: int = 0x1E
TMCC2_DIALOG_CONTROL_EMERGENCY_CONTEXT_DEPENDENT: int = 0x1F
TMCC2_DIALOG_CONTROL_TOWER_CONTEXT_DEPENDENT: int = 0x20
TMCC2_DIALOG_CONTROL_TOWER_DEPARTURE_DENIED: int = 0x22
TMCC2_DIALOG_CONTROL_TOWER_DEPARTURE_GRANTED: int = 0x23
TMCC2_DIALOG_CONTROL_TOWER_DEPARTED: int = 0x24
TMCC2_DIALOG_CONTROL_TOWER_ALL_CLEAR: int = 0x25
TMCC2_DIALOG_CONTROL_TOWER_ARRIVING: int = 0x2D
# noinspection DuplicatedCode
TMCC2_DIALOG_CONTROL_TOWER_ARRIVED: int = 0x2E
TMCC2_DIALOG_CONTROL_TOWER_SHUT_DOWN: int = 0x2F

TMCC2_DIALOG_CONTROL_CONDUCTOR_ALL_ABOARD_A: int = 0x30

TMCC2_DIALOG_CONTROL_ENGINEER_ACK_STANDING_BY: int = 0x31
TMCC2_DIALOG_CONTROL_ENGINEER_ACK_CLEARED_TO_GO: int = 0x32
TMCC2_DIALOG_CONTROL_ENGINEER_ACK_CLEAR_AHEAD: int = 0x33
TMCC2_DIALOG_CONTROL_ENGINEER_ACK_CLEAR_INBOUND: int = 0x34
TMCC2_DIALOG_CONTROL_ENGINEER_ACK_WELCOME_BACK: int = 0x35
TMCC2_DIALOG_CONTROL_ENGINEER_ACK_ID: int = 0x36

TMCC2_DIALOG_CONTROL_ENGINEER_FUEL_LEVEL: int = 0x3D
TMCC2_DIALOG_CONTROL_ENGINEER_FUEL_REFILLED: int = 0x3E
TMCC2_DIALOG_CONTROL_ENGINEER_SPEED: int = 0x3F
TMCC2_DIALOG_CONTROL_ENGINEER_WATER_LEVEL: int = 0x40
TMCC2_DIALOG_CONTROL_ENGINEER_WATER_REFILLED: int = 0x41

TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_OFF: int = 0x50
TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_ON: int = 0x51
TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_CLEAR: int = 0x52
TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_DEPARTED: int = 0x53
TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_TRANSIT: int = 0x54
TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_MAX_SPEED: int = 0x55

TMCC2_DIALOG_CONTROL_CONDUCTOR_NEXT_STOP: int = 0x68
TMCC2_DIALOG_CONTROL_CONDUCTOR_WATCH_YOUR_STEP: int = 0x69
TMCC2_DIALOG_CONTROL_CONDUCTOR_ALL_ABOARD: int = 0x6A
TMCC2_DIALOG_CONTROL_CONDUCTOR_TICKETS_PLEASE: int = 0x6B
TMCC2_DIALOG_CONTROL_CONDUCTOR_PREMATURE_STOP: int = 0x6C

TMCC2_DIALOG_CONTROL_STEWARD_WELCOME_ABOARD: int = 0x6D
TMCC2_DIALOG_CONTROL_STEWARD_FIRST_SEATING: int = 0x6E
TMCC2_DIALOG_CONTROL_STEWARD_SECOND_SEATING: int = 0x6F
TMCC2_DIALOG_CONTROL_STEWARD_LOUNGE_CAR_OPEN: int = 0x70

TMCC2_DIALOG_CONTROL_STATION_ARRIVING: int = 0x71
TMCC2_DIALOG_CONTROL_STATION_ARRIVED: int = 0x72
TMCC2_DIALOG_CONTROL_STATION_BOARDING: int = 0x73
TMCC2_DIALOG_CONTROL_STATION_DEPARTING: int = 0x74

TMCC2_DIALOG_CONTROL_PASSENGER_CAR_STARTUP: int = 0x75
TMCC2_DIALOG_CONTROL_PASSENGER_CAR_SHUTDOWN: int = 0x76

TMCC2_DIALOG_CONTROL_SPECIAL_GUEST_ENABLED: int = 0x7D
TMCC2_DIALOG_CONTROL_SPECIAL_GUEST_DISABLED: int = 0x7E


@unique
class TMCC2RailSoundsDialogControl(TMCC2ParameterEnum):
    CONVENTIONAL_SHUTDOWN = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_CONVENTIONAL_SHUTDOWN)
    SCENE_TWO = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SCENE_TWO)
    SCENE_FIVE = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SCENE_FIVE)
    SCENE_SEVEN = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SCENE_SEVEN)
    SHORT_HORN = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SHORT_HORN)
    TOWER_STARTUP = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_ENGINE_STARTUP)
    ENGINEER_DEPARTURE_DENIED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_DEPARTURE_DENIED)
    ENGINEER_DEPARTURE_GRANTED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_DEPARTURE_GRANTED)
    ENGINEER_DEPARTED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_HAVE_DEPARTED)
    ENGINEER_ALL_CLEAR = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ALL_CLEAR)
    TOWER_SPEED_STOP_HOLD = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_STOP_HOLD)
    TOWER_SPEED_RESTRICTED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_RESTRICTED_SPEED)
    TOWER_SPEED_SLOW = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_SLOW_SPEED)
    TOWER_SPEED_MEDIUM = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_MEDIUM_SPEED)
    TOWER_SPEED_LIMITED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_LIMITED_SPEED)
    TOWER_SPEED_NORMAL = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_NORMAL_SPEED)
    TOWER_SPEED_HIGHBALL = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_HIGHBALL_SPEED)
    ENGINEER_ARRIVING = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ARRIVING_SOON)
    ENGINEER_ARRIVED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_HAVE_ARRIVED)
    ENGINEER_SHUTDOWN = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_SHUTDOWN)
    ENGINEER_ID = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ID)
    ENGINEER_ACK = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ACK)
    ENGINEER_SPEED_STOP_HOLD = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_STOP_SPEED_ACK)
    ENGINEER_SPEED_RESTRICTED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_RESTRICTED_SPEED_ACK)
    ENGINEER_SPEED_SLOW = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_SLOW_SPEED_ACK)
    ENGINEER_SPEED_MEDIUM = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_MEDIUM_SPEED_ACK)
    ENGINEER_SPEED_LIMITED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_LIMITED_SPEED_ACK)
    ENGINEER_SPEED_NORMAL = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_NORMAL_SPEED_ACK)
    ENGINEER_SPEED_HIGHBALL = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_HIGHBALL_SPEED_ACK)
    ENGINEER_CONTEXT_DEPENDENT = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_CONTEXT_DEPENDENT)
    EMERGENCY_CONTEXT_DEPENDENT = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_EMERGENCY_CONTEXT_DEPENDENT)
    TOWER_CONTEXT_DEPENDENT = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_CONTEXT_DEPENDENT)
    TOWER_DEPARTURE_DENIED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_DEPARTURE_DENIED)
    TOWER_DEPARTURE_GRANTED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_DEPARTURE_GRANTED)
    TOWER_DEPARTED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_DEPARTED)
    TOWER_ALL_CLEAR = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_ALL_CLEAR)
    TOWER_ARRIVING = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_ARRIVING)
    TOWER_ARRIVED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_ARRIVED)
    TOWER_SHUTDOWN = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_TOWER_SHUT_DOWN)

    ENGINEER_ACK_STAND_BY = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ACK_STANDING_BY)
    ENGINEER_ACK_CLEARED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ACK_CLEARED_TO_GO)
    ENGINEER_ACK_CLEAR_AHEAD = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ACK_CLEAR_AHEAD)
    ENGINEER_ACK_CLEAR_INBOUND = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ACK_CLEAR_INBOUND)
    ENGINEER_ACK_WELCOME_BACK = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ACK_WELCOME_BACK)
    ENGINEER_ACK_ID = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_ACK_ID)
    ENGINEER_FUEL_LEVEL = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_FUEL_LEVEL)
    ENGINEER_SPEED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_SPEED)
    ENGINEER_FUEL_REFILLED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_FUEL_REFILLED)
    ENGINEER_WATER_LEVEL = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_WATER_LEVEL)
    ENGINEER_WATER_REFILLED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_ENGINEER_WATER_REFILLED)

    CONDUCTOR_NEXT_STOP = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_CONDUCTOR_NEXT_STOP)
    CONDUCTOR_WATCH_YOUR_STEP = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_CONDUCTOR_WATCH_YOUR_STEP)
    CONDUCTOR_ALL_ABOARD = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_CONDUCTOR_ALL_ABOARD)
    CONDUCTOR_TICKETS_PLEASE = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_CONDUCTOR_TICKETS_PLEASE)
    CONDUCTOR_PREMATURE_STOP = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_CONDUCTOR_PREMATURE_STOP)

    STEWARD_WELCOME_ABOARD = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STEWARD_WELCOME_ABOARD)
    STEWARD_FIRST_SEATING = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STEWARD_FIRST_SEATING)
    STEWARD_SECOND_SEATING = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STEWARD_SECOND_SEATING)
    STEWARD_LOUNGE_CAR_OPEN = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STEWARD_LOUNGE_CAR_OPEN)

    STATION_ARRIVING = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STATION_ARRIVING)
    STATION_ARRIVED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STATION_ARRIVED)
    STATION_BOARDING = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STATION_BOARDING)
    STATION_DEPARTING = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_STATION_DEPARTING)

    PASSENGER_CAR_STARTUP = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_PASSENGER_CAR_STARTUP)
    PASSENGER_CAR_SHUTDOWN = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_PASSENGER_CAR_SHUTDOWN)

    SPECIAL_GUEST_ENABLED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SPECIAL_GUEST_ENABLED)
    SPECIAL_GUEST_DISABLED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SPECIAL_GUEST_DISABLED)

    SEQUENCE_OFF = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_OFF)
    SEQUENCE_ON = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_ON)
    SEQUENCE_CLEAR = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_CLEAR)
    SEQUENCE_DEPARTED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_DEPARTED)
    SEQUENCE_TRANSIT = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_TRANSIT)
    SEQUENCE_MAX_SPEED = MultiByteCommandDef(TMCC2_DIALOG_CONTROL_SEQUENCE_CONTROL_MAX_SPEED)


"""
    Word #2 - RailSounds Effects trigger controls (index 0x4)
"""
TMCC2_RS_EFFECTS_PRIME_MOVER_OFF: int = 0x10
TMCC2_RS_EFFECTS_PRIME_MOVER_ON: int = 0x11
TMCC2_RS_EFFECTS_MASTER_VOLUME_DOWN: int = 0x12
TMCC2_RS_EFFECTS_MASTER_VOLUME_UP: int = 0x13
TMCC2_RS_EFFECTS_BLEND_VOLUME_DOWN: int = 0x14
TMCC2_RS_EFFECTS_BLEND_VOLUME_UP: int = 0x15
TMCC2_RS_EFFECTS_CYLINDER_CLEARING_ON: int = 0x20
TMCC2_RS_EFFECTS_CYLINDER_CLEARING_OFF: int = 0x21
TMCC2_RS_EFFECTS_WHEEL_SLIP_TRIGGER: int = 0x22
TMCC2_RS_EFFECTS_STANDBY_WARNING_BELL: int = 0x23
TMCC2_RS_EFFECTS_STANDBY_MODE_DISABLE: int = 0x24
TMCC2_RS_EFFECTS_STANDBY_MODE_ENABLE: int = 0x25
TMCC2_RS_EFFECTS_FORCE_COUPLER_IMPACT_COMPRESS: int = 0x26
TMCC2_RS_EFFECTS_FORCE_COUPLER_IMPACT_STRETCH: int = 0x27
TMCC2_RS_EFFECTS_CIRCUIT_BREAKER_MAIN_LIGHTS: int = 0x28
TMCC2_RS_EFFECTS_CIRCUIT_BREAKER_CAB_LIGHTS: int = 0x29
TMCC2_RS_EFFECTS_CIRCUIT_BREAKER_WORK_LIGHTS: int = 0x2A
TMCC2_RS_EFFECTS_SEQUENCE_CONTROL_OFF: int = 0x2C
TMCC2_RS_EFFECTS_SEQUENCE_CONTROL_ON: int = 0x2D

TMCC2_RS_EFFECTS_RESET_ODOMETER: int = 0x30
TMCC2_RS_EFFECTS_INCREMENT_FUEL_LOAD: int = 0x31

TMCC2_RS_EFFECTS_SOUND_SYSTEM_OFF = TMCC2_SOUND_OFF_COMMAND
TMCC2_RS_EFFECTS_SOUND_SYSTEM_ON = TMCC2_SOUND_ON_COMMAND


@unique
class TMCC2RailSoundsEffectsControl(TMCC2ParameterEnum):
    ADD_FUEL = MultiByteCommandDef(TMCC2_RS_EFFECTS_INCREMENT_FUEL_LOAD)
    BLEND_DOWN = MultiByteCommandDef(TMCC2_RS_EFFECTS_BLEND_VOLUME_DOWN)
    BLEND_UP = MultiByteCommandDef(TMCC2_RS_EFFECTS_BLEND_VOLUME_UP)
    CAB_BREAKER = MultiByteCommandDef(TMCC2_RS_EFFECTS_CIRCUIT_BREAKER_CAB_LIGHTS)
    COUPLER_COMPRESS = MultiByteCommandDef(TMCC2_RS_EFFECTS_FORCE_COUPLER_IMPACT_COMPRESS)
    COUPLER_STRETCH = MultiByteCommandDef(TMCC2_RS_EFFECTS_FORCE_COUPLER_IMPACT_STRETCH)
    CYLINDER_OFF = MultiByteCommandDef(TMCC2_RS_EFFECTS_CYLINDER_CLEARING_OFF)
    CYLINDER_ON = MultiByteCommandDef(TMCC2_RS_EFFECTS_CYLINDER_CLEARING_ON)
    MAIN_BREAKER = MultiByteCommandDef(TMCC2_RS_EFFECTS_CIRCUIT_BREAKER_MAIN_LIGHTS)
    PRIME_OFF = MultiByteCommandDef(TMCC2_RS_EFFECTS_PRIME_MOVER_OFF)
    PRIME_ON = MultiByteCommandDef(TMCC2_RS_EFFECTS_PRIME_MOVER_ON)
    RESET_ODOMETER = MultiByteCommandDef(TMCC2_RS_EFFECTS_RESET_ODOMETER)
    SEQUENCE_CONTROL_OFF = MultiByteCommandDef(TMCC2_RS_EFFECTS_SEQUENCE_CONTROL_OFF)
    SEQUENCE_CONTROL_ON = MultiByteCommandDef(TMCC2_RS_EFFECTS_SEQUENCE_CONTROL_ON)
    STANDBY_BELL = MultiByteCommandDef(TMCC2_RS_EFFECTS_STANDBY_WARNING_BELL)
    STANDBY_DISABLE = MultiByteCommandDef(TMCC2_RS_EFFECTS_STANDBY_MODE_DISABLE)
    STANDBY_ENABLE = MultiByteCommandDef(TMCC2_RS_EFFECTS_STANDBY_MODE_ENABLE)
    VOLUME_DOWN_RS = MultiByteCommandDef(TMCC2_RS_EFFECTS_MASTER_VOLUME_DOWN)
    VOLUME_UP_RS = MultiByteCommandDef(TMCC2_RS_EFFECTS_MASTER_VOLUME_UP)
    WHEEL_SLIP = MultiByteCommandDef(TMCC2_RS_EFFECTS_WHEEL_SLIP_TRIGGER)
    WORK_BREAKER = MultiByteCommandDef(TMCC2_RS_EFFECTS_CIRCUIT_BREAKER_WORK_LIGHTS)


"""
    Word #2 - Sound masking controls (index 0x6)
"""
TMCC2_MASKING_DIALOG_NC_SIGNAL_NC: int = 0x00
TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_NC: int = 0x01
TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_NC: int = 0x02
TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_NC: int = 0x03

TMCC2_MASKING_DIALOG_NC_SIGNAL_PLAY_ALWAYS: int = 0x04
TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_PLAY_ALWAYS: int = 0x05
TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_PLAY_ALWAYS: int = 0x06
TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_PLAY_ALWAYS: int = 0x07

TMCC2_MASKING_DIALOG_NC_SIGNAL_PLAY_NEVER: int = 0x08
TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_PLAY_NEVER: int = 0x09
TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_PLAY_NEVER: int = 0x0A
TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_PLAY_NEVER: int = 0x0B

TMCC2_MASKING_DIALOG_NC_SIGNAL_DEFAULT: int = 0x0C
TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_DEFAULT: int = 0x0D
TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_DEFAULT: int = 0x0E
TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_DEFAULT: int = 0x0F

TMCC2_BRAKE_SQUEAL_DISABLE: int = 0x20
TMCC2_BRAKE_SQUEAL_ENABLE: int = 0x21


@unique
class TMCC2MaskingControl(TMCC2ParameterEnum):
    NC_NC = MultiByteCommandDef(TMCC2_MASKING_DIALOG_NC_SIGNAL_NC)
    ALWAYS_NC = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_NC)
    NEVER_NC = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_NC)
    DEFAULT_NC = MultiByteCommandDef(TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_NC)
    NC_ALWAYS = MultiByteCommandDef(TMCC2_MASKING_DIALOG_NC_SIGNAL_PLAY_ALWAYS)
    ALWAYS_ALWAYS = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_PLAY_ALWAYS)
    NEVER_ALWAYS = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_PLAY_ALWAYS)
    DEFAULT_ALWAYS = MultiByteCommandDef(TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_PLAY_ALWAYS)
    NC_NEVER = MultiByteCommandDef(TMCC2_MASKING_DIALOG_NC_SIGNAL_PLAY_NEVER)
    ALWAYS_NEVER = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_PLAY_NEVER)
    NEVER_NEVER = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_PLAY_NEVER)
    DEFAULT_NEVER = MultiByteCommandDef(TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_PLAY_NEVER)
    NC_DEFAULT = MultiByteCommandDef(TMCC2_MASKING_DIALOG_NC_SIGNAL_DEFAULT)
    PLAY_DEFAULT = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_ALWAYS_SIGNAL_DEFAULT)
    NEVER_DEFAULT = MultiByteCommandDef(TMCC2_MASKING_DIALOG_PLAY_NEVER_SIGNAL_DEFAULT)
    DEFAULT_DEFAULT = MultiByteCommandDef(TMCC2_MASKING_DIALOG_DEFAULT_SIGNAL_DEFAULT)
    BRAKE_SQUEAL_ENABLE = MultiByteCommandDef(TMCC2_BRAKE_SQUEAL_ENABLE)
    BRAKE_SQUEAL_DISABLE = MultiByteCommandDef(TMCC2_BRAKE_SQUEAL_DISABLE)


"""
    Word #2 - Effects controls (index 0xC)
"""
TMCC2_EFFECTS_CONTROL_SMOKE_OFF: int = 0x00
TMCC2_EFFECTS_CONTROL_SMOKE_LOW: int = 0x01
TMCC2_EFFECTS_CONTROL_SMOKE_MEDIUM: int = 0x02
TMCC2_EFFECTS_CONTROL_SMOKE_HIGH: int = 0x03
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_FRONT_UP_CAB2: int = 0x10
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_FRONT_DOWN_CAB2: int = 0x11
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_REAR_UP_CAB2: int = 0x12
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_REAR_DOWN_CAB2: int = 0x13
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_FRONT_UP: int = 0x19
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_FRONT_DOWN: int = 0x18
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_REAR_UP: int = 0x1B
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_REAR_DOWN: int = 0x1A
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_BOTH_UP: int = 0x1F
TMCC2_EFFECTS_CONTROL_PANTOGRAPH_BOTH_DOWN: int = 0x1E
TMCC2_EFFECTS_CONTROL_SUBWAY_LEFT_DOOR_OPEN_CAB2: int = 0x20
TMCC2_EFFECTS_CONTROL_SUBWAY_LEFT_DOOR_CLOSE_CAB2: int = 0x21
TMCC2_EFFECTS_CONTROL_SUBWAY_RIGHT_DOOR_OPEN_CAB2: int = 0x22
TMCC2_EFFECTS_CONTROL_SUBWAY_RIGHT_DOOR_CLOSE_CAB2: int = 0x23
TMCC2_EFFECTS_CONTROL_SUBWAY_LEFT_DOOR_CLOSE: int = 0x28
TMCC2_EFFECTS_CONTROL_SUBWAY_LEFT_DOOR_OPEN: int = 0x29
TMCC2_EFFECTS_CONTROL_SUBWAY_RIGHT_DOOR_CLOSE: int = 0x2A
TMCC2_EFFECTS_CONTROL_SUBWAY_RIGHT_DOOR_OPEN: int = 0x2B
TMCC2_EFFECTS_CONTROL_SUBWAY_BOTH_DOOR_CLOSE: int = 0x2E
TMCC2_EFFECTS_CONTROL_SUBWAY_BOTH_DOOR_OPEN: int = 0x2F
TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_ONE_ON: int = 0x30
TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_ONE_OFF: int = 0x31
TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_TWO_ON: int = 0x32
TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_TWO_OFF: int = 0x33
TMCC2_EFFECTS_CONTROL_STOCK_CAR_LOAD: int = 0x34
TMCC2_EFFECTS_CONTROL_STOCK_CAR_UNLOAD: int = 0x35
TMCC2_EFFECTS_CONTROL_STOCK_CAR_FRED_ON: int = 0x36
TMCC2_EFFECTS_CONTROL_STOCK_CAR_FRED_OFF: int = 0x37
TMCC2_EFFECTS_CONTROL_STOCK_CAR_FLAT_WHEEL_ON: int = 0x38
TMCC2_EFFECTS_CONTROL_STOCK_CAR_FLAT_WHEEL_OFF: int = 0x39
TMCC2_EFFECTS_CONTROL_STOCK_CAR_GAME_ON: int = 0x3A
TMCC2_EFFECTS_CONTROL_STOCK_CAR_GAME_OFF: int = 0x3B
TMCC2_EFFECTS_CONTROL_SCENE_ZERO: int = 0x3C
TMCC2_EFFECTS_CONTROL_SCENE_ONE: int = 0x3D
TMCC2_EFFECTS_CONTROL_SCENE_TWO: int = 0x3E
TMCC2_EFFECTS_CONTROL_SCENE_THREE: int = 0x3F
TMCC2_EFFECTS_CONTROL_COAL_EMPTY: int = 0x50
TMCC2_EFFECTS_CONTROL_COAL_FULL: int = 0x51
TMCC2_EFFECTS_CONTROL_COAL_EMPTYING: int = 0x52
TMCC2_EFFECTS_CONTROL_COAL_FILLING: int = 0x53


@unique
class TMCC2EffectsControl(TMCC2ParameterEnum):
    PANTO_FRONT_DOWN = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_PANTOGRAPH_FRONT_DOWN)
    PANTO_FRONT_UP = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_PANTOGRAPH_FRONT_UP)
    PANTO_REAR_DOWN = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_PANTOGRAPH_REAR_DOWN)
    PANTO_REAR_UP = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_PANTOGRAPH_REAR_UP)
    PANTO_BOTH_DOWN = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_PANTOGRAPH_BOTH_DOWN)
    PANTO_BOTH_UP = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_PANTOGRAPH_BOTH_UP)
    SMOKE_HIGH = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SMOKE_HIGH)
    SMOKE_LOW = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SMOKE_LOW)
    SMOKE_MEDIUM = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SMOKE_MEDIUM)
    SMOKE_OFF = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SMOKE_OFF)
    STOCK_FRED_OFF = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_FRED_OFF)
    STOCK_FRED_ON = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_FRED_ON)
    STOCK_GAME_OFF = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_GAME_OFF)
    STOCK_GAME_ON = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_GAME_ON)
    STOCK_LOAD = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_LOAD)
    STOCK_OPTION_ONE_OFF = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_ONE_OFF)
    STOCK_OPTION_ONE_ON = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_ONE_ON)
    STOCK_OPTION_TWO_OFF = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_TWO_OFF)
    STOCK_OPTION_TWO_ON = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_OPTION_TWO_ON)
    STOCK_UNLOAD = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_UNLOAD)
    STOCK_WHEEL_OFF = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_FLAT_WHEEL_OFF)
    STOCK_WHEEL_ON = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_STOCK_CAR_FLAT_WHEEL_ON)
    SUBWAY_LEFT_DOOR_CLOSE = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SUBWAY_LEFT_DOOR_CLOSE)
    SUBWAY_LEFT_DOOR_OPEN = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SUBWAY_LEFT_DOOR_OPEN)
    SUBWAY_RIGHT_DOOR_CLOSE = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SUBWAY_RIGHT_DOOR_CLOSE)
    SUBWAY_RIGHT_DOOR_OPEN = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SUBWAY_RIGHT_DOOR_OPEN)
    SUBWAY_BOTH_DOOR_CLOSE = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SUBWAY_BOTH_DOOR_CLOSE)
    SUBWAY_BOTH_DOOR_OPEN = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_SUBWAY_BOTH_DOOR_OPEN)
    COAL_EMPTY = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_COAL_EMPTY)
    COAL_FULL = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_COAL_FULL)
    COAL_EMPTYING = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_COAL_EMPTYING)
    COAL_FILLING = MultiByteCommandDef(TMCC2_EFFECTS_CONTROL_COAL_FILLING)


"""
    Word #2 - Lighting controls (index 0xD)
"""
TMCC2_LIGHTING_CONTROL_CAB_LIGHT_AUTO: int = 0xF2
TMCC2_LIGHTING_CONTROL_CAB_LIGHT_OFF: int = 0xF0
TMCC2_LIGHTING_CONTROL_CAB_LIGHT_ON: int = 0xF1
TMCC2_LIGHTING_CONTROL_CAB_LIGHT_TOGGLE: int = 0xF3
TMCC2_LIGHTING_CONTROL_CAR_LIGHT_AUTO: int = 0xFA
TMCC2_LIGHTING_CONTROL_CAR_LIGHT_OFF: int = 0xF8
TMCC2_LIGHTING_CONTROL_CAR_LIGHT_ON: int = 0xF9
TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_OFF: int = 0xC0
TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_OFF_PULSE_ON_WITH_HORN: int = 0xC1
TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_ON: int = 0xC3
TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_ON_PULSE_OFF_WITH_HORN: int = 0xC2
TMCC2_LIGHTING_CONTROL_DOGHOUSE_LIGHT_OFF: int = 0xA0
TMCC2_LIGHTING_CONTROL_DOGHOUSE_LIGHT_ON: int = 0xA1
TMCC2_LIGHTING_CONTROL_GROUND_LIGHT_AUTO: int = 0xD2
TMCC2_LIGHTING_CONTROL_GROUND_LIGHT_OFF: int = 0xD0
TMCC2_LIGHTING_CONTROL_GROUND_LIGHT_ON: int = 0xD1
TMCC2_LIGHTING_CONTROL_HAZARD_LIGHT_OFF: int = 0xB0
TMCC2_LIGHTING_CONTROL_HAZARD_LIGHT_ON: int = 0xB1
TMCC2_LIGHTING_CONTROL_HAZARD_LIGHT_AUTO: int = 0xB2
TMCC2_LIGHTING_CONTROL_LOCO_MARKER_LIGHT_OFF: int = 0xC8
TMCC2_LIGHTING_CONTROL_LOCO_MARKER_LIGHT_ON: int = 0xC9
TMCC2_LIGHTING_CONTROL_LOCO_MARKER_LIGHT_AUTO: int = 0xCA
TMCC2_LIGHTING_CONTROL_MARS_LIGHT_OFF: int = 0xE8
TMCC2_LIGHTING_CONTROL_MARS_LIGHT_ON: int = 0xE9
TMCC2_LIGHTING_CONTROL_RULE_17_AUTO: int = 0xF6
TMCC2_LIGHTING_CONTROL_RULE_17_OFF: int = 0xF4
TMCC2_LIGHTING_CONTROL_RULE_17_ON: int = 0xF5
TMCC2_LIGHTING_CONTROL_STROBE_LIGHT_OFF: int = 0xE0
TMCC2_LIGHTING_CONTROL_STROBE_LIGHT_ON_DOUBLE_FLASH: int = 0xE2
TMCC2_LIGHTING_CONTROL_STROBE_LIGHT_ON_SINGLE_FLASH: int = 0xE1
TMCC2_LIGHTING_CONTROL_TENDER_MARKER_LIGHT_OFF: int = 0xCC
TMCC2_LIGHTING_CONTROL_TENDER_MARKER_LIGHT_ON: int = 0xCD
TMCC2_LIGHTING_CONTROL_WORK_LIGHT_OFF: int = 0xD8
TMCC2_LIGHTING_CONTROL_WORK_LIGHT_ON: int = 0xD9
TMCC2_LIGHTING_CONTROL_WORK_LIGHT_AUTO: int = 0xDA


@unique
class TMCC2LightingControl(TMCC2ParameterEnum):
    CAB_AUTO = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_CAB_LIGHT_AUTO)
    CAB_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_CAB_LIGHT_OFF)
    CAB_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_CAB_LIGHT_ON)
    CAB_TOGGLE = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_CAB_LIGHT_TOGGLE)
    CAR_AUTO = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_CAR_LIGHT_AUTO)
    CAR_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_CAR_LIGHT_OFF)
    CAR_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_CAR_LIGHT_ON)
    DITCH_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_OFF)
    DITCH_OFF_PULSE_ON_WITH_HORN = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_OFF_PULSE_ON_WITH_HORN)
    DITCH_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_ON)
    DITCH_ON_PULSE_OFF_WITH_HORN = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_DITCH_LIGHT_ON_PULSE_OFF_WITH_HORN)
    DOGHOUSE_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_DOGHOUSE_LIGHT_OFF)
    DOGHOUSE_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_DOGHOUSE_LIGHT_ON)
    GROUND_AUTO = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_GROUND_LIGHT_AUTO)
    GROUND_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_GROUND_LIGHT_OFF)
    GROUND_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_GROUND_LIGHT_ON)
    HAZARD_AUTO = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_HAZARD_LIGHT_AUTO)
    HAZARD_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_HAZARD_LIGHT_OFF)
    HAZARD_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_HAZARD_LIGHT_ON)
    LOCO_MARKER_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_LOCO_MARKER_LIGHT_OFF)
    LOCO_MARKER_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_LOCO_MARKER_LIGHT_ON)
    LOCO_MARKER_AUTO = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_LOCO_MARKER_LIGHT_AUTO)
    MARS_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_MARS_LIGHT_OFF)
    MARS_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_MARS_LIGHT_ON)
    RULE_17_AUTO = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_RULE_17_AUTO)
    RULE_17_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_RULE_17_OFF)
    RULE_17_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_RULE_17_ON)
    STROBE_LIGHT_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_STROBE_LIGHT_OFF)
    STROBE_LIGHT_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_STROBE_LIGHT_ON_SINGLE_FLASH)
    STROBE_LIGHT_ON_DOUBLE = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_STROBE_LIGHT_ON_DOUBLE_FLASH)
    TENDER_MARKER_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_TENDER_MARKER_LIGHT_OFF)
    TENDER_MARKER_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_TENDER_MARKER_LIGHT_ON)
    WORK_AUTO = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_WORK_LIGHT_AUTO)
    WORK_OFF = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_WORK_LIGHT_OFF)
    WORK_ON = MultiByteCommandDef(TMCC2_LIGHTING_CONTROL_WORK_LIGHT_ON)


# R4LC Commands; modify Engine eeprom
@unique
class UnitAssignment(Mixins, IntEnum):
    SINGLE_FORWARD = 0b000
    SINGLE_REVERSE = 0b100
    HEAD_FORWARD = 0b001
    HEAD_REVERSE = 0b101
    MIDDLE_FORWARD = 0b010
    MIDDLE_REVERSE = 0b110
    TAIL_FORWARD = 0b011
    TAIL_REVERSE = 0b111


# index values,
TMCC2_SET_ENGINE_ADDRESS_INDEX: int = 0x40
TMCC2_SET_ENGINE_STALL_INDEX: int = 0x41
TMCC2_SET_TRAIN_ADDRESS_INDEX: int = 0x42
TMCC2_SET_TRAIN_UNIT_INDEX: int = 0x43
TMCC2_SET_MAX_SPEED_INDEX: int = 0x44
TMCC2_SET_DIRECTION_INDEX: int = 0x45
TMCC2_SET_FLAGS_INDEX: int = 0x46
TMCC2_SET_CONTROL_INDEX: int = 0x47
TMCC2_SET_ENGINE_TYPE_INDEX: int = 0x48
TMCC2_SET_SPEED_RESOLUTION_INDEX: int = 0x49
TMCC2_SET_VARIABLE_RESOLUTION_INDEX: int = 0x4A
TMCC2_SET_INDIRECT_INDEX: int = 0x4F


@unique
class TMCC2R4LCIndex(Mixins, IntEnum):
    ENGINE_ADDRESS = TMCC2_SET_ENGINE_ADDRESS_INDEX
    ENGINE_STALL = TMCC2_SET_ENGINE_STALL_INDEX
    TRAIN_ADDRESS = TMCC2_SET_TRAIN_ADDRESS_INDEX
    TRAIN_UNIT = TMCC2_SET_TRAIN_UNIT_INDEX
    MAX_SPEED = TMCC2_SET_MAX_SPEED_INDEX
    DIRECTION = TMCC2_SET_DIRECTION_INDEX
    FLAGS = TMCC2_SET_FLAGS_INDEX
    CONTROL = TMCC2_SET_CONTROL_INDEX
    ENGINE_TYPE = TMCC2_SET_ENGINE_TYPE_INDEX
    SPEED_RESOLUTION = TMCC2_SET_SPEED_RESOLUTION_INDEX
    VARIABLE_RESOLUTION = TMCC2_SET_VARIABLE_RESOLUTION_INDEX
    INDIRECT = TMCC2_SET_INDIRECT_INDEX


# I don't like using the same constants for the "data" enums, but for now...
# This will probably stick, as we are really only going to use these to
# echo commands
@unique
class TMCC2R4LCEnum(TMCC2MultiByteEnum):
    ENGINE_ADDRESS = MultiByteCommandDef(TMCC2_SET_ENGINE_ADDRESS_INDEX, d_min=1, d_max=99)
    ENGINE_STALL = MultiByteCommandDef(TMCC2_SET_ENGINE_STALL_INDEX)
    TRAIN_ADDRESS = MultiByteCommandDef(TMCC2_SET_TRAIN_ADDRESS_INDEX, d_max=99)
    TRAIN_UNIT = MultiByteCommandDef(TMCC2_SET_TRAIN_UNIT_INDEX, d_max=7)
    MAX_SPEED = MultiByteCommandDef(TMCC2_SET_MAX_SPEED_INDEX, d_min=32, d_max=200)
    DIRECTION = MultiByteCommandDef(TMCC2_SET_DIRECTION_INDEX, d_max=255)
    FLAGS = MultiByteCommandDef(TMCC2_SET_FLAGS_INDEX, d_max=63)
    CONTROL = MultiByteCommandDef(TMCC2_SET_CONTROL_INDEX, d_max=1)
    ENGINE_TYPE = MultiByteCommandDef(TMCC2_SET_ENGINE_TYPE_INDEX, d_max=7)
    SPEED_RESOLUTION = MultiByteCommandDef(TMCC2_SET_SPEED_RESOLUTION_INDEX, d_max=7)
    VARIABLE_RESOLUTION = MultiByteCommandDef(TMCC2_SET_VARIABLE_RESOLUTION_INDEX)
    INDIRECT = MultiByteCommandDef(TMCC2_SET_INDIRECT_INDEX)


#
# Variable length Multibyte commands
#
class VariableCommandDef(MultiByteCommandDef):
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        command_bits: int,
        num_data_bytes: int,
        data_bytes: List[int] = None,
        d_min: int = 0,
        d_max: int = 0,
    ) -> None:
        super().__init__(command_bits, d_min=d_min, d_max=d_max)
        self._num_data_bytes = num_data_bytes
        self._data_bytes = data_bytes

    @property
    def num_data_bytes(self) -> int:
        return self._num_data_bytes

    @property
    def lsb(self) -> int:
        return 0x00FF & self.bits

    @property
    def msb(self) -> int:
        return (0xFF00 & self.bits) >> 8


TMCC2_VARIABLE_INDEX: int = 0x6F

TMCC2_SET_MASTER_VOLUME: int = 0xB000
TMCC2_SET_BLEND_VOLUME: int = 0xB001
TMCC2_SET_VOLUME_DIRECT: int = 0xB004
TMCC2_DCDS_FACTORY_DEFAULT = 0xF000
TMCC2_DCDS_STORE = 0xF001


class TMCC2VariableEnum(TMCC2MultiByteEnum):
    MASTER_VOLUME = VariableCommandDef(TMCC2_SET_MASTER_VOLUME, 1)
    BLEND_VOLUME = VariableCommandDef(TMCC2_SET_BLEND_VOLUME, 1)
    VOLUME_DIRECT = VariableCommandDef(TMCC2_SET_BLEND_VOLUME, 2)
    FACTORY_DEFAULT = VariableCommandDef(TMCC2_DCDS_FACTORY_DEFAULT, 2, [0xEF, 0xBF])
    STORE = VariableCommandDef(TMCC2_DCDS_FACTORY_DEFAULT, 1)

    @property
    def num_data_bytes(self) -> int:
        return self.value.num_data_bytes



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
from enum import IntEnum, unique
from math import floor
from typing import Dict, List, Tuple

from .constants import PdiCommand, PDI_SOP, PDI_EOP, D4Action
from ..db.comp_data import (
    CompDataMixin,
    CompData,
    CompDataHandler,
    SCOPE_TO_COMP_MAP,
)
from .base3_component import ConsistComponent
from .pdi_req import PdiReq, SCOPE_TO_RECORD_LENGTH, LIONEL_ENGINE_RECORD_LENGTH
from ..db.component_state import ComponentState
from ..protocol.command_def import CommandDefEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope, CONTROL_TYPE, SOUND_TYPE, LOCO_TYPE, LOCO_CLASS, Mixins
from ..protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum

log = logging.getLogger(__name__)

ROUTE_THROW_RATE_MAP: Dict[int, float] = {
    1: 0.00,
    2: 0.25,
    3: 0.50,
    4: 0.75,
    5: 1.00,
    6: 1.25,
    7: 1.50,
    8: 1.75,
    9: 2.00,
}


@unique
class EngineBits(Mixins, IntEnum):
    REVERSE_LINK = 0
    FORWARD_LINK = 1
    ROAD_NAME = 2
    ROAD_NUMBER = 3
    LOCO_TYPE = 4
    CONTROL_TYPE = 5
    SOUND_TYPE = 6
    CLASS_TYPE = 7
    TSDB_LEFT = 8
    TSDB_RIGHT = 9
    SPARE = 10
    SPEED = 11
    RUN_LEVEL = 12
    LABOR_BIAS = 13
    SPEED_LIMIT = 14
    MAX_SPEED = 15
    FUEL_LEVEL = 16
    WATER_LEVEL = 17
    TRAIN_ADDRESS = 18
    TRAIN_POSITION = 19
    SMOKE_LEVEL = 20
    DITCH_LIGHT = 21
    TRAIN_BRAKE = 22
    MOMENTUM = 23


ENGINE_WRITE_MAP = {
    "ABSOLUTE_SPEED": (11, 56, None),
    "STOP_IMMEDIATE": (11, 56, lambda t: 0),
    "RESET": (11, 56, lambda t: 0),
    "DIESEL_RPM": (12, 57, None),
    "ENGINE_LABOR": (13, 58, lambda t: t - 12 if t >= 12 else 20 + t),
    "SMOKE_HIGH": (20, 65, lambda t: 3),
    "SMOKE_MEDIUM": (20, 65, lambda t: 2),
    "SMOKE_LOW": (20, 65, lambda t: 1),
    "SMOKE_OFF": (20, 65, lambda t: 0),
    "DITCH_OFF": (21, 66, lambda t: 0),
    "DITCH_OFF_PULSE_ON_WITH_HORN": (21, 66, lambda t: 1),
    "DITCH_ON_PULSE_OFF_WITH_HORN": (21, 66, lambda t: 2),
    "DITCH_ON": (21, 68, lambda t: 3),
    "TRAIN_BRAKE": (22, 67, lambda t: round(t * 2.143)),
    "MOMENTUM_HIGH": (23, 68, lambda t: 127),
    "MOMENTUM_MEDIUM": (23, 68, lambda t: 63),
    "MOMENTUM_LOW": (23, 68, lambda t: 0),
    "MOMENTUM": (23, 68, lambda t: floor(t * 18.285)),
}


def encode_labor_rpm(rpm: int, labor: int) -> int:
    return rpm | ((labor - 12 if labor >= 12 else 20 + labor) << 3)


def decode_labor_rpm(s: int) -> Tuple[int, int]:
    rpm = s & 0b111
    labor = s >> 3
    labor = labor + 12 if labor <= 19 else labor - 20
    return rpm, labor


RECORD_TYPE_MAP = {
    1: CommandScope.ENGINE,
    2: CommandScope.TRAIN,
    3: CommandScope.ACC,
    5: CommandScope.ROUTE,
    6: CommandScope.SWITCH,
}

SCOPE_TO_RECORD_TYPE_MAP = {s: p for p, s in RECORD_TYPE_MAP.items()}


class BaseReq(PdiReq, CompDataMixin):
    @classmethod
    def update_speed(cls, address: int, speed: int, scope: CommandScope = CommandScope.ENGINE) -> BaseReq:
        if scope == CommandScope.TRAIN:
            return cls(address, PdiCommand.UPDATE_TRAIN_SPEED, flags=0xC2, speed=speed)
        else:
            return cls(address, PdiCommand.UPDATE_ENGINE_SPEED, flags=0xC2, speed=speed)

    @classmethod
    def update_eng(
        cls,
        cmd: CommandDefEnum | CommandReq,
        address: int = None,
        data: int | None = None,
        scope: CommandScope = CommandScope.ENGINE,
    ) -> List[BaseReq] | None:
        from ..db.component_state_store import ComponentStateStore

        pkgs = []
        if isinstance(cmd, CommandReq):
            state = cmd.command
            address = cmd.address
            data = cmd.data
            scope = cmd.scope
            cur_state = ComponentStateStore.build().get_state(scope, address, False)

            # special case numeric commands
            if state.name == "NUMERIC":
                if data in {3, 6}:  # RPM up/down
                    state = TMCC2EngineCommandEnum.DIESEL_RPM
                    if cur_state and cur_state.rpm is not None:
                        cur_rpm = cur_state.rpm
                        cur_labor = cur_state.labor if cur_state.labor else 12
                        if data == 6:  # RPM Down
                            cur_rpm = max(cur_rpm - 1, 0)
                        elif data == 3:  # RPM Up
                            cur_rpm = min(cur_rpm + 1, 7)
                        pkgs.append(CompData.rpm_labor_to_pkg(cur_rpm, cur_labor))
                elif data == 5 and cur_state.speed == 0:  # Shutdown Delayed
                    pkgs.extend(
                        CompData.request_to_updates(
                            CommandReq(
                                TMCC2EngineCommandEnum.SHUTDOWN_DELAYED,
                                address,
                                scope=scope,
                            )
                        )
                    )
                elif data == 0:  # Reset
                    pkgs.extend(
                        CompData.request_to_updates(
                            CommandReq(
                                TMCC2EngineCommandEnum.RESET,
                                address,
                                scope=scope,
                            )
                        )
                    )
        elif isinstance(cmd, CommandDefEnum):
            state = cmd
            cur_state = ComponentStateStore.build().get_state(scope, address, False)
            log.warning("********************** update_eng called with enum not req ***********************")
        else:
            raise ValueError(f"Invalid option: {cmd}")

        # harvest state update pkgs based on command, unless command was numeric 3 or 6
        cmds = []
        pkgs = pkgs if pkgs else CompData.request_to_updates(cmd)
        if pkgs:
            from .d4_req import D4Req

            for pkg in pkgs:
                if 1 <= cmd.address <= 99:
                    cmds.append(
                        BaseReq(
                            cmd.address,
                            pdi_command=PdiCommand.BASE_MEMORY,
                            flags=0xC2,
                            scope=state.scope,
                            start=pkg.offset,
                            data_length=pkg.length,
                            data_bytes=pkg.data_bytes,
                        )
                    )
                else:
                    assert cur_state.record_no != 0xFFFF
                    pdi_cmd = PdiCommand.D4_ENGINE if cmd.scope == CommandScope.ENGINE else PdiCommand.D4_TRAIN
                    cmds.append(
                        D4Req(
                            cur_state.record_no,
                            pdi_cmd,
                            D4Action.UPDATE,
                            start=pkg.offset,
                            data_length=pkg.length,
                            data_bytes=pkg.data_bytes,
                        )
                    )
        elif state.name in ENGINE_WRITE_MAP and cmd.address <= 99:
            bit_pos, offset, scaler = ENGINE_WRITE_MAP[state.name]
            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"State: {state} {data} {bit_pos} {offset} {scaler(data) if scaler else data}")
            value1 = value2 = 0
            if bit_pos <= 15:
                value1 = 1 << bit_pos
            else:
                value2 = 1 << (bit_pos - 16)
            # build data packet
            if scaler:
                data = scaler(data)
            byte_str = bytes()
            pdi_cmd = PdiCommand.BASE_ENGINE if scope == CommandScope.ENGINE else PdiCommand.BASE_TRAIN
            byte_str += pdi_cmd.to_bytes(1, byteorder="big")
            byte_str += address.to_bytes(1, byteorder="big")
            byte_str += (0xC2).to_bytes(1, byteorder="big")
            byte_str += (0x00).to_bytes(2, byteorder="big")  # result byte + spare
            byte_str += value1.to_bytes(2, byteorder="little")
            byte_str += value2.to_bytes(2, byteorder="little")
            # fill the buffer with zeros up to offset point
            byte_str += (0x00).to_bytes(1, byteorder="big") * (offset - len(byte_str))
            # now add data
            byte_str += data.to_bytes(1, byteorder="big")
            # now add SOP, EOP, and Checksum
            byte_str, checksum = cls._calculate_checksum(byte_str)
            byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
            byte_str += checksum
            byte_str += PDI_EOP.to_bytes(1, byteorder="big")
            cmds.append(cls(byte_str))
            log.info(f"Using old-style {pdi_cmd.name} for updates from {cmd}")
        return cmds

    def __init__(
        self,
        data: bytes | int = 0,
        pdi_command: PdiCommand = PdiCommand.BASE,
        flags: int = 2,
        scope: CommandScope = None,
        speed: int = None,
        base_name: str | None = None,
        state: ComponentState = None,
        start: int = 0,
        data_length: int = 0,
        data_bytes: bytes | None = None,
    ) -> None:
        super().__init__(data, pdi_command)
        if self.pdi_command.is_base is False:
            raise ValueError(f"Invalid PDI Base Request: {data}")
        self._status = self._firmware_high = self._firmware_low = None
        self._route_throw_rate = self._name = self._number = None
        self._route_components = None
        self._rev_link = self._fwd_link = None
        self._loco_type = self._control_type = self._sound_type = self._loco_class = None
        self._speed_step = self._speed_limit = self._max_speed = self._labor_bias = None
        self._fuel_level = self._water_level = self._last_train_id = self._train_pos = None
        self._train_brake = None
        self._smoke_level = self._ditch_lights = self._momentum = None
        self._consist_flags = None
        self._consist_comps = []
        self._state = state
        self._valid1 = self._valid2 = None
        self._spare_1 = None
        self._data_length = self._data_bytes = self._start = self._record_type = None
        if isinstance(data, bytes):
            data_len = len(self._data)
            self._record_no = self.tmcc_id = self._data[1] if data_len > 1 else None
            self._flags = self._data[2] if data_len > 2 else None
            self._status = self._data[3] if data_len > 3 else None
            self._spare_1 = self._data[4] if data_len > 4 else None
            self._valid1 = int.from_bytes(self._data[5:7], byteorder="little") if data_len > 5 else None
            self._valid2 = None

            if self.pdi_command in {PdiCommand.BASE_ENGINE, PdiCommand.BASE_TRAIN}:
                self.scope = CommandScope.ENGINE if self.pdi_command == PdiCommand.BASE_ENGINE else CommandScope.TRAIN
                self._valid2 = int.from_bytes(self._data[7:9], byteorder="little") if data_len > 7 else None
                self._rev_link = self._data[9] if data_len > 9 else None
                self._fwd_link = self._data[10] if data_len > 10 else None
                self._name = self.decode_text(self._data[11:44]) if data_len > 11 else None
                self._number = self.decode_text(self._data[44:49]) if data_len > 44 else None
                self._loco_type = self._data[49] if data_len > 49 else None
                self._control_type = self._data[50] if data_len > 50 else None
                self._sound_type = self._data[51] if data_len > 51 else None
                self._loco_class = self._data[52] if data_len > 52 else None
                self._tsdb_left = self._data[53] if data_len > 53 else None
                self._tsdb_right = self._data[54] if data_len > 54 else None
                self._spare = self._data[55] if data_len > 55 else None
                self._speed_step = self._data[56] if data_len > 56 else None
                self._run_level = self._data[57] if data_len > 57 else None
                self._labor_bias = self._data[58] if data_len > 58 else None
                self._speed_limit = self._data[59] if data_len > 59 else None
                self._max_speed = self._data[60] if data_len > 60 else None
                self._fuel_level = self._data[61] if data_len > 61 else None
                self._water_level = self._data[62] if data_len > 62 else None
                self._last_train_id = self._data[63] if data_len > 63 else None
                self._train_pos = self._data[64] if data_len > 64 else None
                self._smoke_level = self._data[65] if data_len > 65 else None
                self._ditch_lights = self._data[66] if data_len > 66 else None
                self._train_brake = self._data[67] if data_len > 67 else None
                self._momentum = self._data[68] if data_len > 68 else None
                if self.pdi_command == PdiCommand.BASE_TRAIN:
                    self._consist_flags = self._data[69] if data_len > 69 else None
                    for i in range(70, 102, 2):
                        if data_len > i:
                            if self._data[i] != 0xFF and self._data[i + 1] != 0xFF:
                                self._consist_comps.insert(0, ConsistComponent(self._data[i], self._data[i + 1]))
                        else:
                            break
            elif self.pdi_command in {PdiCommand.BASE_ACC, PdiCommand.BASE_SWITCH, PdiCommand.BASE_ROUTE}:
                if self.pdi_command == PdiCommand.BASE_ACC:
                    self.scope = CommandScope.ACC
                elif self.pdi_command == PdiCommand.BASE_SWITCH:
                    self.scope = CommandScope.SWITCH
                elif self.pdi_command == PdiCommand.BASE_ROUTE:
                    self.scope = CommandScope.ROUTE
                self._rev_link = self._data[7] if data_len > 7 else None
                self._fwd_link = self._data[8] if data_len > 8 else None
                self._name = self.decode_text(self._data[9:42]) if data_len > 9 else None
                self._number = self.decode_text(self._data[42:47]) if data_len > 42 else None
                if self.scope == CommandScope.ROUTE and self.is_active:
                    # grab the switches/routes in this route
                    if data_len > 47:
                        self._route_components: List[int] = list()
                        for i in range(47, 79, 2):
                            if data_len < i + 2:
                                break
                            tmcc_id = int.from_bytes(self._data[i : i + 2], byteorder="big")
                            if tmcc_id != 0xFFFF and tmcc_id != 0:
                                self._route_components.append(tmcc_id)
            elif self.pdi_command == PdiCommand.BASE:
                self._firmware_high = self._data[7] if data_len > 7 else None
                self._firmware_low = self._data[8] if data_len > 8 else None
                self._route_throw_rate = self.decode_throw_rate(self._data[9]) if data_len > 9 else None
                self._name = self.decode_text(self._data[10:]) if data_len > 10 else None
                self.scope = CommandScope.BASE
            elif self.pdi_command == PdiCommand.BASE_MEMORY:
                record_type = self._data[4] if data_len > 4 else None
                self._scope = RECORD_TYPE_MAP.get(record_type, CommandScope.SYSTEM)
                self._start = int.from_bytes(self._data[5:9], byteorder="little") if data_len > 8 else None
                _ = self._data[9] if data_len > 9 else None  # we assume port is always 2; Database EEProm
                self._data_length = self._data[10] if data_len > 10 else None
                self._data_bytes = self._data[11:] if data_len > 11 else None
                base_record_length = PdiReq.scope_record_length(self.scope)
                if self.data_length == base_record_length and self.scope in {
                    CommandScope.ENGINE,
                    CommandScope.TRAIN,
                    CommandScope.ACC,
                    CommandScope.SWITCH,
                    CommandScope.ROUTE,
                }:
                    self._comp_data = CompData.from_bytes(self._data_bytes, scope=self.scope, tmcc_id=self.tmcc_id)
                    self._name = self._comp_data.road_name
                    self._number = self._comp_data.road_number
                    self._comp_data_record = True  # mark this req as containing a complete CompData record
            elif self.pdi_command in {PdiCommand.UPDATE_ENGINE_SPEED, PdiCommand.UPDATE_TRAIN_SPEED}:
                self._speed = self._data[2] if data_len > 2 else None
                self._valid1 = (1 << EngineBits.SPEED) if data_len > 2 else 0
        else:
            self._record_no = int(data)
            self._flags = flags
            self._speed_step = speed
            self._base_name = base_name
            self._valid1 = 0b1111
            if self.pdi_command == PdiCommand.BASE_MEMORY:
                self._tmcc_id = self.record_no
                self.scope = scope if scope else CommandScope.ENGINE
                self._record_type = SCOPE_TO_RECORD_TYPE_MAP.get(self.scope, 1)
                # if the state contains a valid comp_data, use it as the source
                if start == 0 and data_length == 0 and isinstance(state, CompDataMixin) and state.is_comp_data_record:
                    self._start = 0
                    self._data_length = PdiReq.scope_record_length(self.scope)
                    if state.comp_data is not None:
                        self._data_bytes = state.comp_data.as_bytes()
                else:
                    self._start = start
                    data_length = (
                        data_length
                        if data_length and 0 <= data_length <= PdiReq.scope_record_length(self.scope)
                        else PdiReq.scope_record_length(self.scope)
                    )
                    self._data_length = data_length
                    self._data_bytes = data_bytes
            elif state:
                from .. import TrainState
                from .. import EngineState
                from ..db.base_state import BaseState

                self._status = 0
                self._spare_1 = state.spare_1
                if isinstance(state, BaseState):
                    self._name = state.base_name
                    self._firmware_high = state.firmware_high
                    self._firmware_low = state.firmware_low
                    self._route_throw_rate = state.route_throw_rate
                    self._valid1 = 0b1111
                    self.scope = CommandScope.BASE
                else:
                    self._name = state.road_name
                    self._number = state.road_number
                    self._valid1 = 0b1100
                    self.scope = state.scope
                    if isinstance(state, EngineState):
                        if state.momentum is None:
                            state._momentum = 1
                        if state.train_brake is None:
                            state._train_brake = 0
                        if state.labor is None:
                            state._labor = 12
                        self._valid1 = 0b1111100011111100
                        self._valid2 = 0b11000000
                        self._speed_step = state.speed
                        self._speed_limit = state.speed_limit
                        self._max_speed = state.max_speed
                        self._momentum = min(round(state.momentum * 18.14), 127)
                        self._train_brake = min(round(state.train_brake * 2.143), 15)
                        self._run_level = state.rpm
                        self._labor_bias = state.labor - 12 if state.labor >= 12 else 20 + state.labor
                        self._scope = state.scope
                        self._control_type = state.control_type
                        self._sound_type = state.sound_type
                        self._loco_type = state.engine_type
                        self._loco_class = state.engine_class
                        if isinstance(state, TrainState):
                            self._consist_flags = state.consist_flags
                            self._consist_comps = state.consist_components

    def is_valid(self, field: EngineBits) -> bool:
        if self._valid1 and field.value <= 15:
            return (self._valid1 & (1 << field.value)) != 0
        elif self._valid2 and field.value <= 31:
            return (self._valid2 & (1 << (field.value - 16))) != 0
        return False

    @property
    def as_key(self):
        return self.record_no, self.pdi_command, self.action, self.scope

    @property
    def record_no(self) -> int:
        return self._record_no

    @property
    def last_train_id(self) -> int:
        return self._last_train_id

    @property
    def train_pos(self) -> int:
        return self._train_pos

    @property
    def flags(self) -> int:
        return self._flags

    @property
    def status(self) -> int:
        return self._status

    @property
    def spare_1(self) -> int:
        return self._spare_1

    @property
    def reverse_link(self) -> int:
        return self._rev_link

    @property
    def forward_link(self) -> int:
        return self._fwd_link

    @property
    def valid1(self) -> int:
        return self._valid1

    @property
    def valid2(self) -> int:
        return self._valid2

    @property
    def name(self) -> str:
        return self._name

    @property
    def number(self) -> str:
        return self._number

    @property
    def speed(self) -> int:
        return self._speed_step

    @property
    def max_speed(self) -> int:
        return self._max_speed

    @property
    def speed_limit(self) -> int:
        return self._speed_limit

    @property
    def smoke(self) -> CommandDefEnum | None:
        from .. import SMOKE_LEVEL_MAP

        return SMOKE_LEVEL_MAP.get(self._smoke_level, None)

    @property
    def smoke_level(self) -> int:
        return self._smoke_level

    @property
    def train_brake(self) -> int:
        return self._train_brake

    @property
    def train_brake_tmcc(self) -> int:
        if self.train_brake:
            return min(round(self._train_brake * 0.4667), 7)
        else:
            return self._train_brake

    @property
    def momentum(self) -> int:
        return self._momentum

    @property
    def momentum_tmcc(self) -> int:
        if self.momentum:
            return min(round(self._momentum * 0.05512), 7)
        else:
            return self._momentum

    @property
    def run_level(self) -> int:
        return self._run_level

    @property
    def labor_bias(self) -> int:
        return self._labor_bias

    @property
    def labor_bias_tmcc(self) -> int:
        return self._labor_bias + 12 if self._labor_bias <= 19 else self._labor_bias - 20

    @property
    def is_legacy(self) -> bool:
        return self._control_type == 2

    @property
    def is_tmcc(self) -> bool:
        return self._control_type == 1

    @property
    def is_cab1(self) -> bool:
        return self._control_type == 0

    @property
    def control(self) -> str:
        return CONTROL_TYPE.get(self._control_type, "NA")

    @property
    def control_id(self) -> int:
        return self._control_type

    @property
    def sound(self) -> str:
        return SOUND_TYPE.get(self._sound_type, "NA")

    @property
    def sound_id(self) -> int:
        return self._sound_type

    @property
    def loco_type(self) -> str:
        return LOCO_TYPE.get(self._loco_type, "NA")

    @property
    def loco_type_id(self) -> int:
        return self._loco_type

    @property
    def loco_class(self) -> str:
        return LOCO_CLASS.get(self._loco_class, "NA")

    @property
    def loco_class_id(self) -> int:
        return self._loco_class

    @property
    def consist_components(self) -> List[ConsistComponent]:
        return self._consist_comps

    @property
    def consist_flags(self) -> int:
        return self._consist_flags

    @property
    def components(self) -> List[int]:
        return self._route_components

    @property
    def route_throw_rate(self) -> float:
        return self._route_throw_rate

    @property
    def firmware(self) -> str | None:
        if self._firmware_high and self._firmware_low:
            return f"{self._firmware_high}.{self._firmware_low}"
        return None

    @property
    def is_ack(self) -> bool:
        return self._valid1 is None and self._valid2 is None

    @property
    def is_active(self) -> bool:
        if (
            self.pdi_command == PdiCommand.BASE_MEMORY
            and self.scope in SCOPE_TO_RECORD_LENGTH
            and self.comp_data
            and self.is_comp_data_record is True
        ):
            if (
                self.comp_data.prev_link == 255
                and self.comp_data.next_link == 255
                and not self.name
                and not self.number
            ):
                return False
        if self.pdi_command in {
            PdiCommand.BASE_ENGINE,
            PdiCommand.BASE_TRAIN,
            PdiCommand.BASE_ROUTE,
            PdiCommand.BASE_ACC,
            PdiCommand.BASE_SWITCH,
        }:
            if self.forward_link == 255 and self.reverse_link == 255 and not self.name and not self.number:
                return False
        return True

    @property
    def start(self) -> int:
        return self._start

    @property
    def data_length(self) -> int:
        return self._data_length

    @property
    def data_bytes(self) -> bytes:
        return self._data_bytes

    @property
    def payload(self) -> str:
        f = hex(self.flags) if self.flags is not None else "NA"
        s = self.status if self.status is not None else "NA"
        if self.pdi_command in {PdiCommand.UPDATE_ENGINE_SPEED, PdiCommand.UPDATE_TRAIN_SPEED}:
            scope = "Engine" if self.pdi_command == PdiCommand.UPDATE_ENGINE_SPEED else "Train"
            return f"{scope} #{self.tmcc_id} New Speed: {self.speed}"
        elif self.pdi_command == PdiCommand.BASE_MEMORY and self._start is not None:
            dt = ""
            sc = f"Scope: {self.scope}"
            st = f"Start: {hex(self._start)} Len: {self.data_length}"
            comp_map = SCOPE_TO_COMP_MAP.get(self.scope, None)
            if comp_map:
                tpl = comp_map.get(self.start, None)
                rec_len = SCOPE_TO_RECORD_LENGTH.get(self.scope, None)
            else:
                tpl = None
                rec_len = LIONEL_ENGINE_RECORD_LENGTH
            if self._data_bytes:
                if isinstance(tpl, CompDataHandler) and (tpl.length == self.data_length):
                    dt = f"Data: {tpl.from_bytes(self._data_bytes)}"
                elif self.data_length < rec_len:
                    dt = f"Data: 0x{self._data_bytes.hex(' ')}" if self._data_bytes else "Data: None"
                else:
                    # we have a complete record, don't print it twice
                    pass
            else:
                dt = "Data: None"
            return f"TMCC ID # {self.record_no} {sc} flags: {f} {st} {dt} status: {s} ({self.packet})"
        elif self._valid1 is None and self._valid2 is None:  # ACK received
            return f"Rec # {self.record_no} flags: {f} status: {s} ACK ({self.packet})"
        else:
            v = hex(self.valid1) if self.valid1 is not None else "NA"
            if self.pdi_command in {PdiCommand.BASE_ENGINE, PdiCommand.BASE_TRAIN}:
                tmcc = f"{self.record_no}"
                fwl = f" Fwd: {self._fwd_link}" if self._fwd_link is not None else ""
                rvl = f" Rev: {self._rev_link}" if self._rev_link is not None else ""
                na = f" {self._name}" if self._name is not None else ""
                no = f" {self._number}" if self._number is not None else ""
                ct = f" {self.control}"
                st = f" {self.sound}"
                lt = f" {self.loco_type} ({self._loco_type})"
                lc = f" {self.loco_class} ({self._loco_class})"
                v2 = f" {hex(self._valid2)}" if self._valid2 is not None else ""
                sp = f" SS: {self._speed_step}" if self._speed_step is not None else ""
                sl = f"/{self._speed_limit}" if self._speed_limit is not None else ""
                ms = f"/{self._max_speed}" if self._max_speed is not None else ""
                fl = f" Fuel: {(100.0 * self._fuel_level / 255):.2f}%" if self._fuel_level is not None else ""
                wl = f" Water: {(100.0 * self._water_level / 255):.2f}%" if self._water_level is not None else ""
                rl = f" RL: {self._run_level}" if self._run_level is not None else ""
                el = f" EB: {self._labor_bias}" if self._labor_bias is not None else ""
                sm = f" Smoke: {self._smoke_level}" if self._smoke_level is not None else ""
                m = f" Momentum: {self.momentum} ({self.momentum_tmcc})" if self.momentum is not None else ""
                b = f" Brake: {self._train_brake}" if self._train_brake is not None else ""
                ti = f"Train ID: {self._last_train_id} {self._train_pos} " if self._last_train_id is not None else ""
                if self.pdi_command == PdiCommand.BASE_TRAIN and self._consist_comps:
                    tc = " "
                    for c in self._consist_comps:
                        tc += f"{c} "
                else:
                    tc = ""
                return (
                    f"# {tmcc}{na}{no}{ct}{st}{lt}{lc}{sp}{sl}{ms}{fl}{wl}{rl}{el}{sm}{m}{b}\n"
                    f"{ti}{tc}flags: {f} status: {s} valid: {v}{v2}{fwl}{rvl}\n"
                    f"({self.packet})"
                )
            if self.pdi_command in [PdiCommand.BASE_ACC, PdiCommand.BASE_ROUTE, PdiCommand.BASE_SWITCH]:
                fwl = f" Fwd: {self._fwd_link}" if self._fwd_link is not None else ""
                rvl = f" Rev: {self._rev_link}" if self._rev_link is not None else ""
                na = f" {self._name}" if self._name is not None else ""
                no = f" {self._number}" if self._number is not None else ""
                sw = ""
                if self.pdi_command == PdiCommand.BASE_ROUTE and self._route_components and self.is_active:
                    sw = " Switches:"
                    for comp in self._route_components:
                        state = "out" if (comp & 0x0300) != 0 else "thru"
                        comp &= 0x007F
                        sw += f" {comp} [{state}],"
                return f"# {self.record_no}{na}{no}{sw} flags: {f} status: {s} valid: {v}{fwl}{rvl}\n({self.packet})"
            elif self.pdi_command == PdiCommand.BASE:
                fw = f" V{self._firmware_high}.{self._firmware_low}" if self._firmware_high is not None else ""
                tr = f" Route Throw Rate: {self._route_throw_rate} sec" if self._route_throw_rate is not None else ""
                n = f"{self._name} " if self._name is not None else ""
                return f"{n}Rec # {self.record_no} flags: {f} status: {s} valid: {v}{fw}{tr}\n({self.packet})"
            return f"Rec # {self.record_no} flags: {f} status: {s} valid: {v}\n({self.packet})"

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes
        byte_str += self.record_no.to_bytes(1, byteorder="big")
        if self.pdi_command in {PdiCommand.UPDATE_ENGINE_SPEED, PdiCommand.UPDATE_TRAIN_SPEED}:
            byte_str += self.speed.to_bytes(1, byteorder="little")
        elif self.pdi_command == PdiCommand.BASE_MEMORY:
            byte_str += self.flags.to_bytes(1, byteorder="little")
            byte_str += (0).to_bytes(1, byteorder="little")  # Status
            byte_str += SCOPE_TO_RECORD_TYPE_MAP.get(self.scope, 1).to_bytes(1, byteorder="little")
            byte_str += self._start.to_bytes(4, byteorder="little")
            byte_str += (2).to_bytes(1, byteorder="little")  # database EEProm
            byte_str += self._data_length.to_bytes(1, byteorder="little")
            if self._data_bytes:
                byte_str += self._data_bytes
        elif self._state:
            byte_str += self.flags.to_bytes(1, byteorder="little")
            byte_str += self.status.to_bytes(1, byteorder="little")
            byte_str += (self.spare_1 if self.spare_1 else 0).to_bytes(1, byteorder="big")  # spare
            byte_str += self._valid1.to_bytes(2, byteorder="little")
            if self._valid2 is not None:
                byte_str += self._valid2.to_bytes(2, byteorder="little")
            if self.pdi_command == PdiCommand.BASE:
                byte_str += self._firmware_high.to_bytes(1, byteorder="little")
                byte_str += self._firmware_low.to_bytes(1, byteorder="little")
                byte_str += self.encode_throw_rate(self.route_throw_rate).to_bytes(1, byteorder="little")
                byte_str += self.encode_text(self.name, 33)
            else:
                byte_str += (101).to_bytes(1, byteorder="little")  # rev link
                byte_str += (101).to_bytes(1, byteorder="little")  # fwd link
                byte_str += self.encode_text(self.name, 33)
                byte_str += self.encode_text(self.number, 5)
                if self.pdi_command in {PdiCommand.BASE_ENGINE, PdiCommand.BASE_TRAIN}:
                    byte_str += self._loco_type.to_bytes(1, byteorder="little")  # loco type
                    byte_str += self._control_type.to_bytes(1, byteorder="little")
                    byte_str += self._sound_type.to_bytes(1, byteorder="little")
                    byte_str += self._loco_class.to_bytes(1, byteorder="little")
                    byte_str += (0).to_bytes(3, byteorder="little")  # 3 misc fields
                    byte_str += self._speed_step.to_bytes(1, byteorder="little")
                    byte_str += self._run_level.to_bytes(1, byteorder="little")
                    byte_str += self._labor_bias.to_bytes(1, byteorder="little")  # labor bias
                    byte_str += self._speed_limit.to_bytes(1, byteorder="little")
                    byte_str += self._max_speed.to_bytes(1, byteorder="little")
                    byte_str += (0).to_bytes(6, byteorder="little")
                    byte_str += self._train_brake.to_bytes(1, byteorder="little")
                    byte_str += self._momentum.to_bytes(1, byteorder="little")
                    if self.pdi_command == PdiCommand.BASE_TRAIN and self._consist_comps:
                        ff = (0xFF).to_bytes(1, byteorder="little")
                        cf = self._consist_flags if self._consist_flags else 0
                        byte_str += cf.to_bytes(1, byteorder="little")
                        byte_str += 2 * ff
                        i = 15
                        for c in reversed(self.consist_components):
                            byte_str += c.flags.to_bytes(1, byteorder="little")
                            byte_str += c.tmcc_id.to_bytes(1, byteorder="little")
                            i -= 1
                        byte_str += (max(i, 0) * 2) * ff
                elif self.pdi_command == PdiCommand.BASE_ROUTE:
                    for sw in self._route_components:
                        byte_str += sw.to_bytes(2, byteorder="big")
        else:
            byte_str += self.flags.to_bytes(1, byteorder="big")
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str

    @staticmethod
    def decode_throw_rate(data: int) -> float | None:
        if data in ROUTE_THROW_RATE_MAP:
            return ROUTE_THROW_RATE_MAP[data]
        else:
            return None

    @staticmethod
    def encode_throw_rate(data: float) -> int | None:
        if data is not None:
            for key, value in ROUTE_THROW_RATE_MAP.items():
                if value <= data < value + 0.25:
                    return key
        return 9



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#
#  SPDX-License-Identifier: LPGL
#
#
import pytest
from src.pytrain.gpio.utils.text_buffer import TextBuffer


class TestTextBuffer:
    def test_initialization(self):
        buffer = TextBuffer(rows=5, cols=10, auto_update=False)
        assert buffer.rows == 5
        assert buffer.cols == 10
        assert not buffer.auto_update
        assert len(buffer) == 0
        assert buffer.cursor_pos == (0, 0)

    def test_repr_empty(self):
        buffer = TextBuffer(rows=3)
        assert repr(buffer) == ""

    def test_set_and_get_item(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Hello"
        buffer[1] = "World"
        assert buffer[0] == "Hello"
        assert buffer[1] == "World"

    def test_set_item_out_of_bounds(self):
        buffer = TextBuffer(rows=3)
        with pytest.raises(IndexError):
            buffer[5] = "Out of bounds"

    def test_delete_item(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Row 1"
        buffer[1] = "Row 2"
        del buffer[0]
        assert len(buffer) == 1
        assert buffer[0] == "Row 2"

    def test_insert_item(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Row 1"
        buffer.insert(0, "Inserted Row")
        assert len(buffer) == 2
        assert buffer[0] == "Inserted Row"
        assert buffer[1] == "Row 1"

    def test_insert_out_of_bounds(self):
        buffer = TextBuffer(rows=3)
        with pytest.raises(IndexError):
            buffer.insert(5, "Out of bounds")

    def test_reverse_buffer(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Row 1"
        buffer[1] = "Row 2"
        buffer.reverse()
        assert buffer[0] == "Row 2"
        assert buffer[1] == "Row 1"

    def test_add_row(self):
        buffer = TextBuffer(rows=3)
        assert buffer.add("New Row")
        assert len(buffer) == 1
        assert buffer[0] == "New Row"

    def test_add_row_exceeds_limit(self):
        buffer = TextBuffer(rows=1)
        buffer.add("Row 1")
        assert not buffer.add("Row 2")

    def test_contains(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Hello"
        assert "Hello" in buffer
        assert "World" not in buffer

    def test_count(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Hello"
        buffer[1] = "Hello"
        assert buffer.count("Hello") == 2
        assert buffer.count("World") == 0

    def test_clear_buffer(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Row 1"
        buffer[1] = "Row 2"
        buffer.clear()
        assert len(buffer) == 0
        assert buffer.cursor_pos == (0, 0)

    def test_auto_update_property(self):
        buffer = TextBuffer(auto_update=True)
        assert buffer.auto_update
        buffer.auto_update = False
        assert not buffer.auto_update

    def test_cursor_pos_setting(self):
        buffer = TextBuffer(rows=3, cols=5)
        buffer.cursor_pos = (1, 4)
        assert buffer.cursor_pos == (1, 4)

    def test_cursor_pos_invalid(self):
        buffer = TextBuffer(rows=3, cols=5)
        with pytest.raises(AttributeError):
            buffer.cursor_pos = (5, 4)

    def test_changed_rows(self):
        buffer = TextBuffer(rows=3)
        buffer[0] = "Row 1"
        buffer[1] = "Row 2"
        assert buffer.changed_rows == [0, 1]
        assert buffer.changed_rows == []

    def test_write_char(self):
        buffer = TextBuffer(rows=3, cols=5)
        buffer.write_chr(65, (0, 0))
        assert buffer[0] == "A"

    def test_write_string(self):
        buffer = TextBuffer(rows=3, cols=10)
        buffer.write("Hello", (1, 0))
        assert buffer[1] == "Hello"

    def test_write_out_of_bounds(self):
        buffer = TextBuffer(rows=3, cols=10)
        with pytest.raises(AttributeError):
            buffer.write("Overflow", (5, 0))



from __future__ import annotations

from .constants import PdiCommand, Asc2Action, PDI_SOP, PDI_EOP
from .lcs_req import LcsReq
from ..protocol.constants import CommandScope


class Asc2Req(LcsReq):
    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = PdiCommand.ASC2_GET,
        action: Asc2Action = Asc2Action.CONFIG,
        ident: int | None = None,
        error: bool = False,
        mode: int = None,
        debug: int = None,
        delay: float = None,
        values: int = None,
        valids: int = None,
        time: float = None,
        sub_id: int = None,
    ) -> None:
        super().__init__(data, pdi_command, action, ident, error)
        if isinstance(data, bytes):
            self._action = Asc2Action(self._action_byte)
            data_len = len(self._data)
            if self._action == Asc2Action.CONFIG:
                self._debug = self._data[4] if data_len > 4 else None
                self._mode = self._data[7] if data_len > 7 else None
                self._delay = self._data[8] / 100.0 if data_len > 8 else None
            else:
                self._mode = self._debug = self._delay = None

            if self._action == Asc2Action.CONTROL1:
                self._values = self._data[3] if data_len > 3 else None
                self._time = self._data[4] / 100.0 if data_len > 4 else None
                self._valids = self._sub_id = self._thru = None
                self.scope = CommandScope.ACC
            elif self._action in [Asc2Action.CONTROL2, Asc2Action.CONTROL3]:
                self._values = self._data[3] if data_len > 3 else None
                self._valids = self._data[4] if data_len > 4 else None
                self._time = self._sub_id = self._thru = None
                self.scope = CommandScope.ACC
            elif self._action == Asc2Action.CONTROL4:
                self._values = self._data[3] if data_len > 3 else None
                self._time = self._data[4] / 100.0 if data_len > 4 else None
                self._valids = self._sub_id = None
                self._thru = self._values == 0 if self._values is not None else None
                self.scope = CommandScope.SWITCH
            elif self._action == Asc2Action.CONTROL5:
                self._values = self._data[3] if data_len > 3 else None
                self._valids = self._sub_id = self._time = None
                self._thru = self._values == 0 if self._values is not None else None
                self.scope = CommandScope.SWITCH
            else:
                self._values = self._valids = self._time = self._sub_id = self._thru = None
        else:
            self._action = action
            self._mode = mode
            self._debug = debug
            self._delay: float = delay
            self._values = values
            self._valids = valids
            self._time = time
            self._sub_id = sub_id

    @property
    def mode(self) -> int | None:
        return self._mode

    @property
    def debug(self) -> int | None:
        return self._debug

    @property
    def delay(self) -> float | None:
        return self._delay

    @property
    def values(self) -> int | None:
        return self._values

    @property
    def state(self) -> int | None:
        return self._values

    @property
    def valids(self) -> int | None:
        return self._valids

    @property
    def sub_id(self) -> int | None:
        return self._sub_id

    @property
    def time(self) -> float | None:
        return self._time

    @property
    def is_thru(self) -> bool | None:
        return self._thru

    @property
    def is_out(self) -> bool | None:
        return not self._thru if self._thru is not None else None

    @property
    def payload(self) -> str | None:
        if self.is_error:
            return super().payload
        if self.pdi_command != PdiCommand.ASC2_GET:
            if self.action == Asc2Action.CONFIG:
                return f"Mode: {self.mode} Debug: {self.debug} Delay: {self.delay} ({self.packet})"
            elif self.action == Asc2Action.CONTROL1:
                time = f" for {self.time:.2f} s" if self.values == 1 and self.time else ""
                return f"Relay: {'ON' if self.values == 1 else 'OFF'}{time} ({self.packet})"
            elif self.action == Asc2Action.CONTROL2:
                return f"Relays: {self.values} Valids: {self.valids} ({self.packet})"
            elif self.action == Asc2Action.CONTROL3:
                if self.pdi_command == PdiCommand.ASC2_RX:
                    return f"Relays: {self.values} Valids: {self.valids} ({self.packet})"
                elif self.pdi_command == PdiCommand.ASC2_SET:
                    return f"Sub ID: {self.sub_id} Time: {self.time} ({self.packet})"
            elif self.action == Asc2Action.CONTROL4:
                if self.pdi_command == PdiCommand.ASC2_SET:
                    return f"{'THRU' if self.values == 0 else 'OUT'} Time: {self.time} ({self.packet})"
                elif self.pdi_command == PdiCommand.ASC2_RX:
                    return f"{'THRU' if self.values == 0 else 'OUT'} ({self.packet})"
            elif self.action == Asc2Action.CONTROL5:
                return f"{'THRU' if self.values == 0 else 'OUT'} ({self.packet})"
        return super().payload

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes
        byte_str += self.tmcc_id.to_bytes(1, byteorder="big")
        byte_str += self.action.as_bytes
        if self._action == Asc2Action.CONFIG:
            if self.pdi_command != PdiCommand.ASC2_GET:
                debug = self.debug if self.debug is not None else 0
                delay = int(round((self.delay * 100))) if self.delay is not None else 0
                byte_str += self.tmcc_id.to_bytes(1, byteorder="big")  # allows board to be renumbered
                byte_str += debug.to_bytes(1, byteorder="big")
                byte_str += (0x0000).to_bytes(2, byteorder="big")
                byte_str += self.mode.to_bytes(1, byteorder="big")
                byte_str += delay.to_bytes(1, byteorder="big")
        elif self._action == Asc2Action.CONTROL1:
            if self.pdi_command != PdiCommand.ASC2_GET:
                values = self.values if self.values is not None else 0
                time = min(int(round((self.time * 100))), 255) if self.time is not None else 0
                byte_str += values.to_bytes(1, byteorder="big")
                byte_str += time.to_bytes(1, byteorder="big")
        elif self._action == Asc2Action.CONTROL2:
            if self.pdi_command != PdiCommand.ASC2_GET:
                values = self.values if self.values is not None else 0
                valids = self.values if self.valids is not None else 0
                byte_str += values.to_bytes(1, byteorder="big")
                byte_str += valids.to_bytes(1, byteorder="big")
        elif self._action == Asc2Action.CONTROL3:
            if self.pdi_command != PdiCommand.ASC2_GET:
                if self.pdi_command == PdiCommand.ASC2_SET:
                    values = self.sub_id if self.values is not None else 1
                    time = min(int(round((self.time * 100))), 255) if self.time is not None else 0
                    if time == 1:
                        time = 0
                    valids = time
                else:
                    values = self.values if self.valids is not None else 0
                    valids = self.valids if self.valids is not None else 0
                byte_str += values.to_bytes(1, byteorder="big")
                byte_str += valids.to_bytes(1, byteorder="big")
        elif self._action == Asc2Action.CONTROL4:
            if self.pdi_command != PdiCommand.ASC2_GET:
                values = self.values if self.values is not None else 0
                time = min(int(round((self.time * 100))), 255) if self.time is not None else 0
                if time == 1:
                    time = 0
                byte_str += values.to_bytes(1, byteorder="big")
                byte_str += time.to_bytes(1, byteorder="big")
        elif self._action == Asc2Action.CONTROL5:
            if self.pdi_command != PdiCommand.ASC2_GET:
                values = self.values if self.values is not None else 0
                byte_str += values.to_bytes(1, byteorder="big")
        elif self._action == Asc2Action.IDENTIFY:
            if self.pdi_command == PdiCommand.ASC2_SET:
                byte_str += (self.ident if self.ident is not None else 0).to_bytes(1, byteorder="big")
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str



from itertools import repeat
from threading import RLock
from typing import Tuple

from gpiozero import GPIODeviceClosed, SourceMixin
from gpiozero.threads import GPIOThread

from .i2c_device import I2CDevice
from .mcp23017 import OUTPUT, Mcp23017Factory


class LEDI2C(I2CDevice, SourceMixin):
    def __init__(
        self,
        pin: int | Tuple[int, int],
        i2c_address: int = 0x23,
        cathode: bool = True,
        initial_value: bool = False,
        pin_factory=None,
    ):
        # i2c buttons use the MCP 23017 i2c dio board, which supports 16 pins and interrupts
        if isinstance(pin, tuple):
            if len(pin) > 1 and pin[1]:
                i2c_address = pin[1]
            if len(pin) > 2 and pin[2] in {True, False}:
                cathode = pin[2]
            if len(pin) > 3 and pin[3] in {True, False}:
                initial_value = pin[3]
            if len(pin) > 0 and pin[0] in range(16):
                pin = pin[0]
            else:
                raise ValueError(f"Invalid pin specifier: {pin}")
        self._dio_pin = pin
        self._cathode = cathode
        self._mcp_23017 = Mcp23017Factory.build(
            address=i2c_address,
            pin=pin,
            client=self,
        )
        # initialize the gpiozero device
        super().__init__(pin_factory=pin_factory)
        self.source_delay = 0.2
        self._controller = None
        self._blink_thread = None
        self._lock = RLock()

        # configure the Mcp23017 pin to the appropriate mode
        self._mcp_23017.set_pin_mode(pin, OUTPUT)

        # If cathode is connected to GPIO,
        if cathode is False:
            self._mcp_23017.set_value(pin, 1)  #  turns the pin off
        if initial_value is not None:
            self.value = 1 if initial_value is True else 0

    def __repr__(self):
        # noinspection PyBroadException
        try:
            return f"<{self.__class__.__name__} object on pin {self.pin} is_active={self.is_active}>"
        except Exception:
            return super().__repr__()

    def _signal_event(self, active: bool) -> None:
        # Noop for LEDs
        pass

    @property
    def bounce_time(self) -> float | None:
        return None

    @property
    def pin(self) -> int:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C LED is closed or uninitialized")
        return self._dio_pin

    def close(self) -> None:
        with self._lock:
            try:
                self._stop_blink()
                # in edge cases where constructor fails, _mcp_23017 property may not exist
                if hasattr(self, "_mcp_23017") and self._mcp_23017 is not None:
                    self.value = 0
                    self.source = None
                    self._mcp_23017.deregister_client(self)
                    Mcp23017Factory.close(self._mcp_23017, self._dio_pin)
                self._mcp_23017 = None
            finally:
                super().close()

    @property
    def closed(self) -> bool:
        return self._mcp_23017 is None

    @property
    def i2c_address(self) -> int:
        return self._mcp_23017.address

    @property
    def value(self) -> int:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C LED is closed or uninitialized")
        value = self._mcp_23017.value(self._dio_pin)
        if self._cathode is False:
            value = 1 if value == 0 else 0
        return value

    @value.setter
    def value(self, value: int) -> None:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C LED is closed or uninitialized")
        if self._cathode is False:
            value = 1 if value == 0 else 0
        self._mcp_23017.set_value(self._dio_pin, value)

    def on(self):
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C LED is closed or uninitialized")
        self._stop_blink()
        self.value = 1

    def off(self):
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C LED is closed or uninitialized")
        self._stop_blink()
        self.value = 0

    def toggle(self):
        """
        Reverse the state of the device. If it's on, turn it off; if it's off,
        turn it on.
        """
        with self._lock:
            if self.is_active:
                self.off()
            else:
                self.on()

    @property
    def is_active(self) -> bool:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C LED is closed or uninitialized")
        return self.value == 1

    @SourceMixin.source.setter  # override setter
    def source(self, value) -> None:
        with self._lock:
            if self._mcp_23017 is None:
                raise GPIODeviceClosed("I2C LED is closed or uninitialized")
            self._stop_blink()
            SourceMixin.source.fset(self, value)

    def blink(self, on_time=1, off_time=1, n=None, background=True):
        """
        Make the device turn on and off repeatedly.

        :param float on_time:
            Number of seconds on. Defaults to 1 second.

        :param float off_time:
            Number of seconds off. Defaults to 1 second.

        :type n: int or None
        :param n:
            Number of times to blink; :data:`None` (the default) means forever.

        :param bool background:
            If :data:`True` (the default), start a background thread to
            continue blinking and return immediately. If :data:`False`, only
            return when the blink is finished (warning: the default value of
            *n* will result in this method never returning).
        """
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C LED is closed or uninitialized")
        self.source = None
        self._stop_blink()
        self._blink_thread = GPIOThread(self._blink_device, (on_time, off_time, n))
        self._blink_thread.start()
        if not background:
            self._blink_thread.join()
            self._blink_thread = None

    # noinspection PyProtectedMember
    def _stop_blink(self):
        if getattr(self, "_controller", None):
            self._controller._stop_blink(self)
        self._controller = None
        if getattr(self, "_blink_thread", None):
            self._blink_thread.stop()
        self._blink_thread = None

    def _blink_device(self, on_time, off_time, n):
        iterable = repeat(0) if n is None else repeat(0, n)
        for _ in iterable:
            self.value = 1
            if self._blink_thread.stopping.wait(on_time):
                break
            self.value = 0
            if self._blink_thread.stopping.wait(off_time):
                break


LEDI2C.is_lit = LEDI2C.is_active



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
import socketserver
import threading
from threading import Event
from typing import cast

from ..comm.comm_buffer import CommBuffer, CommBufferProxy
from ..comm.command_listener import CommandListener, Subscriber, Topic
from ..pdi.constants import PDI_SOP, PDI_EOP, PDI_STF
from ..protocol.command_def import CommandDefEnum

log = logging.getLogger(__name__)


class ClientStateListener(threading.Thread):
    _instance = None
    _lock = threading.RLock()

    @classmethod
    def build(cls) -> ClientStateListener:
        return ClientStateListener()

    @classmethod
    def listen_for(
        cls, listener: Subscriber, channel: Topic, address: int = None, command: CommandDefEnum = None, data: int = None
    ):
        cls.build().subscribe(listener, channel, address, command, data)

    def __init__(self) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        from .. import PROGRAM_NAME

        super().__init__(daemon=True, name=f"{PROGRAM_NAME} ComponentStateListener")
        self._tmcc_listener = CommandListener.build(ser2_receiver=False, base3_receiver=False)
        from ..pdi.pdi_listener import PdiListener

        self._pdi_listener = PdiListener.build(build_base3_reader=False)
        self._tmcc_buffer = cast(CommBufferProxy, CommBuffer.build())
        self._port = self._tmcc_buffer.server_port()
        self._is_running = True
        self._ev = Event()
        self.start()

        # wait for socket server to be up and running
        self._ev.wait()
        self._tmcc_buffer.register(self.port)  # register this client with server to receive updates

        # See if this client needs an upgrade. If it does, the actual upgrade
        # will be done by the PyTrain main program
        if self.update_client_if_needed(False) is True:
            return

        self._tmcc_buffer.sync_state()  # request initial state from server
        self._tmcc_buffer.start_heart_beat()

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in the system
        """
        with cls._lock:
            if ClientStateListener._instance is None:
                ClientStateListener._instance = super(ClientStateListener, cls).__new__(cls)
                ClientStateListener._instance._initialized = False
            return ClientStateListener._instance

    @property
    def port(self) -> int:
        return self._port

    def run(self) -> None:
        while self._is_running:
            try:
                # noinspection PyTypeChecker
                with socketserver.TCPServer(("", self._port), ClientStateHandler) as server:
                    # inform main thread server is running on a valid port
                    self._ev.set()
                    server.serve_forever()
            except OSError as oe:
                if oe.errno in {48, 98}:
                    self._port += 1
                else:
                    raise oe

    def update_client_if_needed(self, do_upgrade=True) -> bool:
        from .. import get_version, get_version_tuple

        # wait for client registration to happen and for the server to tell the client its version
        # if the version of the server is newer, we want to update the client
        self._tmcc_buffer.server_version_available().wait()
        server_version = self._tmcc_buffer.server_version
        client_version = get_version_tuple()  # only interested in major and minor version
        if server_version is None or server_version > client_version:
            if do_upgrade is True:
                cv = f"{get_version()}"
                sv = f" --> v{server_version[0]}.{server_version[1]}.{server_version[2]}" if server_version else ""
                log.info(f"Client needs update: {cv}{sv}")
            return True
        return False

    def shutdown(self) -> None:
        self._is_running = False
        # noinspection PyTypeChecker
        with socketserver.TCPServer(("", self._port), ClientStateHandler) as server:
            server.shutdown()

    def offer(self, data: bytes) -> None:
        # look at first byte to determine handler
        from ..pdi.constants import PDI_SOP

        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"ClientStateListener Offered: {data.hex(' ')}")
        if data and data[0] == PDI_SOP:
            self._pdi_listener.offer(data)
        else:
            self._tmcc_listener.offer(data)

    def subscribe(
        self,
        listener: Subscriber,
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ) -> None:
        self._tmcc_listener.subscribe(listener, channel, address, command, data)
        self._pdi_listener.subscribe(listener, channel, address)

    def unsubscribe(
        self,
        listener: Subscriber,
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ):
        self._tmcc_listener.unsubscribe(listener, channel, address, command, data)
        self._pdi_listener.unsubscribe(listener, channel, address)


class ClientStateHandler(socketserver.BaseRequestHandler):
    def handle(self):
        csl = ClientStateListener.build()
        byte_stream = bytes()
        while True:
            data = self.request.recv(512)
            if data:
                byte_stream += data
                self.request.sendall(str.encode("ack"))
            else:
                break
        # the byte stream could be a combo of PDI AND TMCC commands; we don't
        # want to duplicate all the byte stuffing code, but if the first byte
        # is PDI_SOP, look for a PDI_EOP and just send that portion
        while byte_stream:
            command_bytes = byte_stream
            if byte_stream[0] == PDI_SOP:
                if PDI_EOP in byte_stream:
                    eop_index = byte_stream.index(PDI_EOP)
                    while (eop_index - 1) >= 0 and byte_stream[eop_index - 1] == PDI_STF:
                        if PDI_EOP in byte_stream[eop_index + 1 :]:
                            eop_index = byte_stream.index(PDI_EOP, eop_index + 1)
                        else:
                            break
                    command_bytes = byte_stream[0 : eop_index + 1]
                    byte_stream = byte_stream[eop_index + 1 :]
            else:
                byte_stream = bytes()
            csl.offer(command_bytes)



import threading
import time
from collections import defaultdict
from queue import Queue
from typing import Any

# noinspection PyPackageRequirements
import pytest

# noinspection PyProtectedMember
from src.pytrain.comm.command_listener import CommandListener, CommandDispatcher, Message, Channel
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import DEFAULT_QUEUE_SIZE, BROADCAST_TOPIC, CommandScope
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1HaltCommandEnum, TMCC1SwitchCommandEnum
from src.pytrain.protocol.tmcc2.tmcc2_constants import (
    TMCC2EngineCommandEnum,
    TMCC2RouteCommandEnum,
    TMCC2HaltCommandEnum,
)
from test.test_base import TestBase

CALLBACK_DICT = {}


# noinspection PyTypeChecker
@pytest.fixture(autouse=True)
def run_before_and_after_tests(tmpdir) -> None:
    """
    Fixture to execute asserts before and after a test is run
    """
    # Setup: fill with any logic you want

    yield  # this is where the testing happens

    # Teardown : fill with any logic you want
    CALLBACK_DICT.clear()
    if CommandListener.is_built():
        CommandListener().shutdown()
    assert CommandListener.is_built() is False

    if CommandDispatcher.is_built():
        CommandDispatcher().shutdown()
    assert CommandDispatcher.is_built() is False


class TestCommandDispatcher(TestBase):
    def __call__(self, message: Message) -> None:
        self.register_callback(BROADCAST_TOPIC, message)

    def switch_topic(self, message: Message) -> None:
        self.register_callback(CommandScope.SWITCH, message)

    def engine_topic(self, message: Message) -> None:
        self.register_callback(CommandScope.ENGINE, message)

    def engine_13_topic(self, message: Message) -> None:
        self.register_callback((CommandScope.ENGINE, 13), message)

    def engine_22_ring_bell_topic(self, message: Message) -> None:
        self.register_callback((CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL), message)

    @staticmethod
    def register_callback(topic: Any, message: Message) -> None:
        if topic in CALLBACK_DICT:
            CALLBACK_DICT[topic].append(message)
        else:
            CALLBACK_DICT[topic] = [message]

    def teardown_method(self, test_method):
        super().teardown_method(test_method)

    def test_command_dispatcher_singleton(self) -> None:
        assert CommandDispatcher.is_built() is False
        dispatcher = CommandDispatcher()
        time.sleep(0.25)
        assert dispatcher.is_built
        assert dispatcher.is_running() is True
        assert isinstance(dispatcher, CommandDispatcher)
        assert dispatcher is CommandDispatcher()
        assert dispatcher.is_alive()
        assert dispatcher.broadcasts_enabled is False
        assert isinstance(dispatcher._channels, defaultdict)
        assert not dispatcher._channels
        assert dispatcher.daemon is True
        assert dispatcher._queue is not None  # queue should exist
        assert dispatcher._queue.maxsize == DEFAULT_QUEUE_SIZE
        assert isinstance(dispatcher._queue, Queue)
        assert dispatcher._queue.empty()  # should be empty
        assert dispatcher._cv is not None
        assert isinstance(dispatcher._cv, threading.Condition)

        # shutdown should clear singleton, forcing a new one to be created
        assert dispatcher._instance is not None
        dispatcher.shutdown()
        assert dispatcher._instance is None
        assert CommandDispatcher._instance is None
        assert dispatcher.is_built() is False
        assert dispatcher.is_running() is False
        assert CommandDispatcher.is_built() is False
        assert dispatcher != CommandDispatcher()
        CommandDispatcher().shutdown()
        assert CommandDispatcher._instance is None

        # creation of a CommandListener should create a command dispatcher
        listener = CommandListener()
        dispatcher = CommandDispatcher()
        assert listener._dispatcher is dispatcher
        assert dispatcher.is_running() is True

        # shutting down the listener should shut down the dispatcher
        listener.shutdown()
        assert listener.is_running() is False

    def test_command_dispatcher_run(self) -> None:
        dispatcher = CommandDispatcher()
        # add elements to the queue and make sure they appear in deque in the correct order
        ring_req = CommandReq.build(TMCC2EngineCommandEnum.RING_BELL, 20)
        halt_req = CommandReq.build(TMCC1HaltCommandEnum.HALT)
        with dispatcher._cv:
            dispatcher.offer(ring_req)
            dispatcher.offer(halt_req)
            dispatcher.offer(ring_req)
            # queue should contain 3 entries
            assert dispatcher._queue.empty() is False
            assert dispatcher._queue.qsize() == 3
            req = dispatcher._queue.get_nowait()
            dispatcher._queue.task_done()
            assert req is not None
            assert req == ring_req

            req = dispatcher._queue.get_nowait()
            dispatcher._queue.task_done()
            assert req is not None
            assert req == halt_req

            req = dispatcher._queue.get_nowait()
            dispatcher._queue.task_done()
            assert req is not None
            assert req == ring_req
            assert dispatcher._queue.empty()

    def test_channel_class(self):
        channel = Channel()
        assert channel is not None
        assert channel.subscribers is not None
        assert not channel.subscribers
        assert isinstance(channel.subscribers, set)

        # add this class as a subscriber
        channel.subscribe(self)
        assert channel.subscribers
        assert len(channel.subscribers) == 1

        # add self again, should not change subscriber count
        channel.subscribe(self)
        assert len(channel.subscribers) == 1

        # unsubscribe
        channel.unsubscribe(self)
        assert len(channel.subscribers) == 0

        # resubscribe and try publishing
        channel.subscribe(self)
        channel.publish("ABC")
        assert len(CALLBACK_DICT) == 1
        assert CALLBACK_DICT[BROADCAST_TOPIC] == ["ABC"]

        # remove subscriber and republish; should be no exception
        channel.unsubscribe(self)
        CALLBACK_DICT.clear()
        channel.publish("ABC")
        assert len(CALLBACK_DICT) == 0

    def test_publish_all(self) -> None:
        # create dispatcher and add some channels
        dispatcher = CommandDispatcher()
        assert dispatcher.broadcasts_enabled is False
        dispatcher.subscribe_any(self)
        assert dispatcher.broadcasts_enabled is True

        dispatcher.unsubscribe_any(self)
        assert dispatcher.broadcasts_enabled is False

        # test publish_all
        dispatcher.subscribe_any(self)
        assert dispatcher.broadcasts_enabled is True
        ring_req = CommandReq.build(TMCC2EngineCommandEnum.RING_BELL, 2)
        dispatcher.offer(ring_req)
        time.sleep(0.1)
        dispatcher.shutdown()
        dispatcher.join()
        assert dispatcher.is_running() is False
        assert len(CALLBACK_DICT) == 1
        assert CALLBACK_DICT[BROADCAST_TOPIC] == [ring_req]

    # noinspection DuplicatedCode
    def test_publish(self) -> None:
        # create dispatcher and add some channels
        dispatcher = CommandDispatcher()
        assert dispatcher.broadcasts_enabled is False

        # register callbacks
        assert len(dispatcher._channels) == 0
        dispatcher.subscribe(self.switch_topic, CommandScope.SWITCH)
        dispatcher.subscribe(self.engine_topic, CommandScope.ENGINE)
        dispatcher.subscribe(self.engine_13_topic, CommandScope.ENGINE, 13)
        dispatcher.subscribe(self.engine_22_ring_bell_topic, CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL)
        assert dispatcher.broadcasts_enabled is False
        assert len(dispatcher._channels) == 4

        # offer an Engine Req, should only trigger one listener
        ring_req = CommandReq.build(TMCC2EngineCommandEnum.RING_BELL, 3)
        assert ring_req.address == 3
        dispatcher.offer(ring_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        assert len(dispatcher._channels) == 4
        # listener should have triggered one exception
        assert len(CALLBACK_DICT) == 1
        assert CALLBACK_DICT[CommandScope.ENGINE] == [ring_req]

        # retest for eng request to engine 13
        CALLBACK_DICT.clear()
        ring_req = CommandReq.build(TMCC2EngineCommandEnum.RING_BELL, 13)
        assert ring_req.address == 13
        dispatcher.offer(ring_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        # listener should have triggered one exception
        assert len(CALLBACK_DICT) == 2
        assert CALLBACK_DICT[(CommandScope.ENGINE, 13)] == [ring_req]
        assert CALLBACK_DICT[CommandScope.ENGINE] == [ring_req]

        # retest for eng request to engine 22
        CALLBACK_DICT.clear()
        ring_req = CommandReq.build(TMCC2EngineCommandEnum.RING_BELL, 22)
        assert ring_req.address == 22
        dispatcher.offer(ring_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        # listener should have triggered one exception
        assert len(CALLBACK_DICT) == 2
        assert CALLBACK_DICT[(CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL)] == [ring_req]
        assert CALLBACK_DICT[CommandScope.ENGINE] == [ring_req]

        # retest for route request, should generate no callbacks
        CALLBACK_DICT.clear()
        rte_req = CommandReq.build(TMCC2RouteCommandEnum.FIRE, 13)
        dispatcher.offer(rte_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        assert len(CALLBACK_DICT) == 0

        # unsubscribe engine 22 and fire request, callback should not be invoked
        dispatcher.unsubscribe(
            self.engine_22_ring_bell_topic, CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL
        )
        assert len(dispatcher._channels) == 3
        ring_req = CommandReq.build(TMCC2EngineCommandEnum.RING_BELL, 22)
        assert ring_req.address == 22
        dispatcher.offer(ring_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        # listener should have triggered one exception
        assert len(CALLBACK_DICT) == 1
        assert CALLBACK_DICT[CommandScope.ENGINE] == [ring_req]

        # retest switch, should only be one callback
        CALLBACK_DICT.clear()
        sw_req = CommandReq.build(TMCC1SwitchCommandEnum.OUT, 22)
        assert sw_req.address == 22
        dispatcher.offer(sw_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        # listener should have triggered one callback
        assert len(CALLBACK_DICT) == 1
        assert CALLBACK_DICT[CommandScope.SWITCH] == [sw_req]

    # noinspection DuplicatedCode
    def test_publish_halt(self) -> None:
        # create dispatcher and add some channels
        dispatcher = CommandDispatcher()
        assert dispatcher.broadcasts_enabled is False

        # register callbacks
        assert len(dispatcher._channels) == 0
        dispatcher.subscribe(self.switch_topic, CommandScope.SWITCH)
        dispatcher.subscribe(self.engine_topic, CommandScope.ENGINE)
        dispatcher.subscribe(self.engine_13_topic, CommandScope.ENGINE, 13)
        dispatcher.subscribe(self.engine_22_ring_bell_topic, CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL)
        assert dispatcher.broadcasts_enabled is False
        assert len(dispatcher._channels) == 4

        # send a halt command; should be received by all listeners
        halt_req = CommandReq.build(TMCC1HaltCommandEnum.HALT)
        dispatcher.offer(halt_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        assert len(dispatcher._channels) == 4
        # listener should have triggered 4 exception
        assert len(CALLBACK_DICT) == 4
        assert CALLBACK_DICT[CommandScope.ENGINE] == [halt_req]
        assert CALLBACK_DICT[(CommandScope.ENGINE, 13)] == [halt_req]
        assert CALLBACK_DICT[(CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL)] == [halt_req]
        assert CALLBACK_DICT[CommandScope.SWITCH] == [halt_req]

    # noinspection DuplicatedCode
    def test_publish_system_halt(self) -> None:
        #  Test receipt of SYSTEM_HALT
        dispatcher = CommandDispatcher()
        assert dispatcher.broadcasts_enabled is False

        # register callbacks
        assert len(dispatcher._channels) == 0
        dispatcher.subscribe(self.switch_topic, CommandScope.SWITCH)
        dispatcher.subscribe(self.engine_topic, CommandScope.ENGINE)
        dispatcher.subscribe(self.engine_13_topic, CommandScope.ENGINE, 13)
        dispatcher.subscribe(self.engine_22_ring_bell_topic, CommandScope.TRAIN, 22, TMCC2EngineCommandEnum.RING_BELL)
        assert dispatcher.broadcasts_enabled is False
        assert len(dispatcher._channels) == 4

        # send a halt command; should be received by all listeners
        sys_halt_req = CommandReq.build(TMCC2HaltCommandEnum.HALT)
        dispatcher.offer(sys_halt_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        assert len(dispatcher._channels) == 4
        # listener should have triggered engine and train channels
        assert len(CALLBACK_DICT) == 3
        assert CALLBACK_DICT[CommandScope.ENGINE] == [sys_halt_req]
        assert CALLBACK_DICT[(CommandScope.ENGINE, 13)] == [sys_halt_req]
        assert CALLBACK_DICT[(CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL)] == [sys_halt_req]

        # enable broadcasts and retest
        CALLBACK_DICT.clear()
        dispatcher.subscribe_any(self)
        assert len(dispatcher._channels) == 5
        dispatcher.offer(sys_halt_req)
        time.sleep(0.05)
        assert dispatcher.is_running() is True
        assert dispatcher.broadcasts_enabled
        assert len(CALLBACK_DICT) == 4
        assert CALLBACK_DICT[BROADCAST_TOPIC] == [sys_halt_req]
        assert CALLBACK_DICT[CommandScope.ENGINE] == [sys_halt_req]
        assert CALLBACK_DICT[(CommandScope.ENGINE, 13)] == [sys_halt_req]
        assert CALLBACK_DICT[(CommandScope.ENGINE, 22, TMCC2EngineCommandEnum.RING_BELL)] == [sys_halt_req]
